"""Env creation utilities."""
import functools
import importlib
import inspect
from typing import Callable, Mapping

import gym
from gym.wrappers import TimeLimit
from ray.tune.registry import ENV_CREATOR, _global_registry

from .wrappers import AddRelativeTimestep, SinglePrecision


def has_env_creator(env_id: str) -> bool:
    "Whether and environment with the given id is in the global registry."
    return _global_registry.contains(ENV_CREATOR, env_id)


def get_env_creator(env_id: str) -> Callable[[dict], gym.Env]:
    """Return the environment creator funtion for the given environment id."""
    if not _global_registry.contains(ENV_CREATOR, env_id):
        raise ValueError(f"Environment id {env_id} not registered in Tune")
    return _global_registry.get(ENV_CREATOR, env_id)


def wrap_if_needed(env_creator):
    """Wraps an env creator function to handle time limit configurations."""

    @functools.wraps(env_creator)
    def wrapped(config):
        tmp = config.copy()
        time_limit = [tmp.pop(k, None) for k in ("time_aware", "max_episode_steps")]
        single_precision = tmp.pop("single_precision", False)
        env = env_creator(tmp)
        env = wrap_time_limit(env, *time_limit)
        if single_precision:
            env = SinglePrecision(env)
        return env

    return wrapped


def wrap_time_limit(env, time_aware, max_episode_steps):
    """Add or update an environment's time limit

    Arguments:
        env (gym.Env): a gym environment instance
        time_aware (bool): whether to append the relative timestep to the observation.
        max_episode_steps (int): the maximum number of timesteps in a single episode

    Returns:
        A wrapped environment with the desired time limit
    """
    assert not time_aware or max_episode_steps, "Time-aware envs must specify a horizon"

    if max_episode_steps:
        env_, has_timelimit = env, False
        while hasattr(env_, "env"):
            if isinstance(env_, TimeLimit):
                has_timelimit = True
                break
            env_ = env_.env

        if has_timelimit:
            # pylint:disable=protected-access
            env_._max_episode_steps = max_episode_steps
            # pylint: enable=protected-access
        else:
            env = TimeLimit(env, max_episode_steps=max_episode_steps)

    if time_aware:
        env = AddRelativeTimestep(env)

    return env


def get_env_parameters(env_id: str) -> Mapping[str, inspect.Parameter]:
    """Return an unwrapped environment's constructor parameters.

    Args:
        env_id: The environment name registered in Gym

    Returns:
        A mapping from parameter names to Parameter objects
    """
    env_spec = gym.spec(env_id)
    mod_str, cls_str = env_spec.entry_point.split(":")
    mod = importlib.import_module(mod_str)
    env_cls = getattr(mod, cls_str)

    return inspect.signature(env_cls).parameters



"""CLI for rolling out trained policies."""
import json

import click

from .utils import initialize_raylab


@click.command()
@click.argument(
    "checkpoint",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
)
@click.option(
    "--agent", required=True, default=None, help="Name of the trainable class to run."
)
@click.option(
    "--env",
    default=None,
    help="Name of the environment to interact with. "
    "Optional; can be inferred from config.",
    show_default=True,
)
@click.option(
    "--eval-config/--train-config",
    default=True,
    help="Whether to use the evaluation_config for the agent.",
    show_default=True,
)
@click.option(
    "--render/--no-render",
    default=False,
    help="Whether to render each episode.",
    show_default=True,
)
@click.option(
    "--video-dir",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    default=None,
    help="Specifies the directory into which videos of all episode "
    "rollouts will be stored.",
)
@click.option(
    "--steps",
    default=10000,
    help="Number of timesteps to roll out (overwritten by --episodes).",
    show_default=True,
)
@click.option(
    "--episodes",
    default=0,
    help="Number of complete episodes to roll out (overrides --steps).",
    show_default=True,
)
@click.option(
    "--out",
    type=click.Path(file_okay=True, dir_okay=False, resolve_path=True),
    default=None,
    help="Output filename.",
)
@click.option(
    "--config",
    default="{}",
    type=json.loads,
    help="Algorithm-specific configuration (e.g. env, hyperparams). "
    "Gets merged with loaded configuration from checkpoint file and "
    "`evaluation_config` settings therein.",
)
@click.option(
    "--save-info/--ignore-info",
    default=False,
    help="Save the info field generated by the step() method, "
    "as well as the action, observations, rewards and done fields.",
    show_default=True,
)
@click.option(
    "--shelve/--no-shelve",
    default=False,
    help="Save rollouts into a python shelf file (will save each episode "
    "as it is generated). An output filename must be set using --out.",
    show_default=True,
)
@click.option(
    "--track-progress/--no-progress",
    default=False,
    help="Write progress to a temporary file (updated "
    "after each episode). An output filename must be set using --out; "
    "the progress file will live in the same folder.",
    show_default=True,
)
@initialize_raylab
def rollout(checkpoint, agent, env, eval_config, **rollout_kwargs):
    """Wrap `rllib rollout` with customized options."""
    # pylint:disable=too-many-locals
    import ray
    from ray.rllib.rollout import RolloutSaver
    from ray.rllib.rollout import rollout as rllib_rollout

    from raylab.utils.checkpoints import get_agent_from_checkpoint

    ray.init()
    agent = get_agent_from_checkpoint(
        checkpoint,
        agent,
        env,
        use_eval_config=eval_config,
        config_overrides=rollout_kwargs["config"],
    )

    with RolloutSaver(
        rollout_kwargs["out"],
        rollout_kwargs["shelve"],
        write_update_file=rollout_kwargs["track_progress"],
        target_steps=rollout_kwargs["steps"],
        target_episodes=rollout_kwargs["episodes"],
        save_info=rollout_kwargs["save_info"],
    ) as saver:
        rllib_rollout(
            agent,
            env,
            rollout_kwargs["steps"],
            rollout_kwargs["episodes"],
            saver,
            not rollout_kwargs["render"],
            rollout_kwargs["video_dir"],
        )



# pylint:disable=missing-docstring,invalid-name
from typing import List, Tuple, Union

import gym
import numpy as np
import torch
from torch import Tensor

DEFAULT_CONFIG = {
    "MAX_RES_CAP": [100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0, 100.0],
    "UPPER_BOUND": [62.5, 62.5, 62.5, 62.5, 62.5, 62.5, 62.5, 62.5],
    "LOWER_BOUND": [37.5, 37.5, 37.5, 37.5, 37.5, 37.5, 37.5, 37.5],
    "RAIN_SHAPE": [0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25],
    "RAIN_SCALE": [20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0],
    "DOWNSTREAM": [
        [False, True, False, False, False, False, False, False],
        [False, False, True, False, False, False, False, False],
        [False, False, False, True, False, False, False, False],
        [False, False, False, False, True, False, False, False],
        [False, False, False, False, False, True, False, False],
        [False, False, False, False, False, False, True, False],
        [False, False, False, False, False, False, False, True],
        [False, False, False, False, False, False, False, False],
    ],
    "SINK_RES": [False, False, False, False, False, False, False, True],
    "MAX_WATER_EVAP_FRAC_PER_TIME_UNIT": 0.05,
    "LOW_PENALTY": [-0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5],
    "HIGH_PENALTY": [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
    "init": {"rlevel": [5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0]},
    "horizon": 40,
}


class ReservoirEnv(gym.Env):

    metadata = {"render.modes": ["human"]}

    def __init__(self, config=None):
        self._config = self._convert_to_tensor({**DEFAULT_CONFIG, **(config or {})})

        self._num_reservoirs = len(self._config["init"]["rlevel"])

        self.action_space = gym.spaces.Box(
            low=np.array([0.0] * self._num_reservoirs, dtype=np.float32),
            high=np.array([1.0] * self._num_reservoirs, dtype=np.float32),
        )

        self.observation_space = gym.spaces.Box(
            low=np.array([0.0] * (self._num_reservoirs + 1), dtype=np.float32),
            high=np.array([np.inf] * (self._num_reservoirs) + [1.0], dtype=np.float32),
        )

        self._state = None
        self.reset()

        self._horizon = self._config["horizon"]

    @staticmethod
    def _convert_to_tensor(config: dict) -> dict:
        conf_ = config.copy()
        for key in [
            "RAIN_SHAPE",
            "RAIN_SCALE",
            "DOWNSTREAM",
            "MAX_RES_CAP",
            "LOWER_BOUND",
            "UPPER_BOUND",
            "LOW_PENALTY",
            "HIGH_PENALTY",
        ]:
            conf_[key] = torch.as_tensor(conf_[key])
        return conf_

    def reset(self):
        self._state = np.array(self._config["init"]["rlevel"] + [0.0])
        return self._state

    @torch.no_grad()
    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, dict]:
        state, action = map(torch.as_tensor, (self._state, action))
        next_state, _ = self.dynamics_fn(state, action)
        reward = self.reward_fn(state, action, next_state).item()
        self._state = next_state.numpy()
        return self._state, reward, self._terminal(), {}

    def dynamics_fn(
        self, state: Tensor, action: Tensor, sample_shape: List[int] = ()
    ) -> Tuple[Tensor, Tensor]:
        # pylint:disable=missing-docstring
        state, time = self._unpack_state(state)
        rain, logp = self._rainfall(state, sample_shape)
        action = torch.as_tensor(action) * state
        next_state = self._rlevel(state, action, rain)
        time = self._step_time(time)
        time = time.expand_as(next_state[..., -1:])
        return torch.cat([next_state, time], dim=-1), logp

    def _rlevel(self, rlevel: Tensor, action: Tensor, rain: Tensor) -> Tensor:
        MIN_RES_CAP = torch.zeros(self._num_reservoirs)
        outflow = torch.as_tensor(action)
        rlevel = rlevel + rain - self._evaporated(rlevel)
        rlevel = (
            rlevel
            + self._inflow(rlevel, action)
            - outflow
            - self._overflow(rlevel, action)
        )
        return torch.max(MIN_RES_CAP, rlevel)

    def _rainfall(
        self, rlevel: Tensor, sample_shape: List[int] = ()
    ) -> Tuple[Tensor, Tensor]:
        concentration = self._config["RAIN_SHAPE"].expand_as(rlevel)
        rate = 1.0 / self._config["RAIN_SCALE"].expand_as(rlevel)
        dist = torch.distributions.Independent(
            torch.distributions.Gamma(concentration, rate), reinterpreted_batch_ndims=1
        )
        sample = dist.rsample(sample_shape)
        logp = dist.log_prob(sample.detach())
        return sample, logp

    def _inflow(self, rlevel: Tensor, action: Tensor) -> Tensor:
        DOWNSTREAM = self._config["DOWNSTREAM"].float()
        overflow = self._overflow(rlevel, action)
        outflow = action
        return (overflow + outflow).matmul(DOWNSTREAM.t())

    def _overflow(self, rlevel: Tensor, action: Tensor) -> Tensor:
        MIN_RES_CAP = torch.zeros(self._num_reservoirs)
        MAX_RES_CAP = self._config["MAX_RES_CAP"]
        outflow = torch.as_tensor(action)
        return torch.max(MIN_RES_CAP, rlevel - outflow - MAX_RES_CAP)

    def _evaporated(self, rlevel: Tensor) -> Tensor:
        # EVAP_PER_TIME_UNIT = self._config["MAX_WATER_EVAP_FRAC_PER_TIME_UNIT"]
        MAX_RES_CAP = self._config["MAX_RES_CAP"]
        # return (
        #     EVAP_PER_TIME_UNIT
        #     * torch.log(1.0 + rlevel)
        #     * (rlevel ** 2)
        #     / (MAX_RES_CAP ** 2)
        # )
        return 1 / 2 * torch.sin(rlevel / MAX_RES_CAP) * rlevel

    def _step_time(self, time: Tensor) -> Tensor:
        timestep = torch.round(self._horizon * time)
        return torch.clamp((timestep + 1) / self._horizon, 0, 1)

    def reward_fn(self, state: Tensor, action: Tensor, next_state: Tensor) -> Tensor:
        # pylint:disable=unused-argument,missing-docstring
        rlevel, _ = self._unpack_state(next_state)

        LOWER_BOUND = self._config["LOWER_BOUND"]
        UPPER_BOUND = self._config["UPPER_BOUND"]

        LOW_PENALTY = self._config["LOW_PENALTY"]
        HIGH_PENALTY = self._config["HIGH_PENALTY"]

        # 0.01 * abs(rlevel - (lower_bound + upper_bound) / 2) + high_penalty * max(
        #     0.0, rlevel - upper_bound
        # ) + low_penalty * max(0.0, lower_bound - rlevel)

        mean_capacity_deviation = -0.01 * torch.abs(
            rlevel - (LOWER_BOUND + UPPER_BOUND) / 2
        )
        overflow_penalty = HIGH_PENALTY * torch.max(
            torch.zeros_like(rlevel), rlevel - UPPER_BOUND
        )
        underflow_penalty = LOW_PENALTY * torch.max(
            torch.zeros_like(rlevel), LOWER_BOUND - rlevel
        )

        penalty = mean_capacity_deviation + overflow_penalty + underflow_penalty
        return penalty.sum(dim=-1)

    def _terminal(self):
        _, time = self._unpack_state(self._state)
        return time.item() >= 1.0

    def termination_fn(
        self, state: Tensor, action: Tensor, next_state: Tensor
    ) -> Tensor:
        # pylint:disable=unused-argument,missing-docstring
        _, time = self._unpack_state(next_state)
        return time[..., 0] >= 1.0

    @staticmethod
    def _unpack_state(state: Union[np.ndarray, Tensor]) -> Tuple[Tensor, Tensor]:
        obs = torch.as_tensor(state[..., :-1], dtype=torch.float32)
        time = torch.as_tensor(state[..., -1:], dtype=torch.float32)
        return obs, time

    def render(self, mode="human"):
        pass



# pylint:disable=missing-docstring,invalid-name
import gym
import numpy as np
import torch

DEFAULT_CONFIG = {
    "ADJ": [[False, True, True], [False, False, True], [False, False, False]],
    "ADJ_OUTSIDE": [True, True, False],
    "ADJ_HALL": [True, False, True],
    "R_OUTSIDE": [4.0, 4.0, 4.0],
    "R_HALL": [2.0, 2.0, 2.0],
    "R_WALL": [[1.5, 1.5, 1.5], [1.5, 1.5, 1.5], [1.5, 1.5, 1.5]],
    "IS_ROOM": [True, True, True],
    "CAP": [80.0, 80.0, 80.0],
    "CAP_AIR": 1.006,
    "COST_AIR": 1.0,
    "TIME_DELTA": 1.0,
    "TEMP_AIR": 40.0,
    "TEMP_UP": [23.5, 23.5, 23.5],
    "TEMP_LOW": [20.0, 20.0, 20.0],
    "PENALTY": 20000.0,
    "AIR_MAX": [10.0, 10.0, 10.0],
    "TEMP_OUTSIDE_MEAN": [6.0, 6.0, 6.0],
    "TEMP_OUTSIDE_VARIANCE": [1.0, 1.0, 1.0],
    "TEMP_HALL_MEAN": [10.0, 10.0, 10.0],
    "TEMP_HALL_VARIANCE": [1.0, 1.0, 1.0],
    "init": {"temp": [10.0, 10.0, 10.0]},
    "horizon": 40,
}


class HVACEnv(gym.Env):

    metadata = {"render.modes": ["human"]}

    def __init__(self, config=None):
        self._config = {**DEFAULT_CONFIG, **(config or {})}

        self._num_rooms = len(self._config["init"]["temp"])

        self.observation_space = gym.spaces.Box(
            low=np.array([-np.inf] * self._num_rooms + [0.0], dtype=np.float32),
            high=np.array([np.inf] * self._num_rooms + [1.0], dtype=np.float32),
        )
        self.action_space = gym.spaces.Box(
            low=np.array([0.0] * self._num_rooms, dtype=np.float32),
            high=np.array([1.0] * self._num_rooms, dtype=np.float32),
        )

        self._horizon = self._config["horizon"]
        self._state = None
        self.reset()

    def reset(self):
        self._state = np.array(self._config["init"]["temp"] + [0.0])
        return self._state

    @property
    def temp(self):
        obs, _ = self._unpack_state(self._state)
        return torch.as_tensor(obs, dtype=torch.float32)

    @torch.no_grad()
    def step(self, action):
        state, action = map(torch.as_tensor, (self._state, action))
        next_state, _ = self.transition_fn(state, action)
        reward = self.reward_fn(state, action, next_state).item()
        self._state = next_state.numpy()
        return self._state, reward, self._terminal(), {}

    def transition_fn(self, state, action, sample_shape=()):
        # pylint:disable=missing-docstring
        state, time = self._unpack_state(state)

        AIR_MAX = torch.as_tensor(self._config["AIR_MAX"])
        action = torch.as_tensor(action) * AIR_MAX

        temp_hall, logp_temp_hall = self._temp_hall(sample_shape)
        temp_outside, logp_temp_outside = self._temp_outside(sample_shape)

        next_state = self._temp(action, temp_outside, temp_hall)
        logp = logp_temp_hall + logp_temp_outside
        time = self._step_time(time)
        time = time.expand_as(next_state[..., -1:])
        return torch.cat([next_state, time], dim=-1), logp

    def _temp_hall(self, sample_shape=()):
        TEMP_HALL_MEAN = torch.as_tensor(self._config["TEMP_HALL_MEAN"])
        TEMP_HALL_VARIANCE = torch.sqrt(
            torch.as_tensor(self._config["TEMP_HALL_VARIANCE"])
        )
        dist = torch.distributions.Normal(TEMP_HALL_MEAN, TEMP_HALL_VARIANCE)
        sample = dist.rsample(sample_shape)
        logp = dist.log_prob(sample.detach())
        return sample, logp

    def _temp_outside(self, sample_shape=()):
        TEMP_OUTSIDE_MEAN = torch.as_tensor(self._config["TEMP_OUTSIDE_MEAN"])
        TEMP_OUTSIDE_VARIANCE = torch.sqrt(
            torch.as_tensor(self._config["TEMP_OUTSIDE_VARIANCE"])
        )
        dist = torch.distributions.Normal(TEMP_OUTSIDE_MEAN, TEMP_OUTSIDE_VARIANCE)
        sample = dist.rsample(sample_shape)
        logp = dist.log_prob(sample.detach())
        return sample, logp

    def _temp(self, action, temp_outside, temp_hall):  # pylint:disable=too-many-locals
        air = action

        TIME_DELTA = torch.as_tensor(self._config["TIME_DELTA"])

        CAP = torch.as_tensor(self._config["CAP"])
        CAP_AIR = torch.as_tensor(self._config["CAP_AIR"])
        TEMP_AIR = torch.as_tensor(self._config["TEMP_AIR"])

        IS_ROOM = torch.as_tensor(self._config["IS_ROOM"])

        ADJ = torch.as_tensor(self._config["ADJ"])
        ADJ_OUTSIDE = torch.as_tensor(self._config["ADJ_OUTSIDE"])
        ADJ_HALL = torch.as_tensor(self._config["ADJ_HALL"])

        R_OUTSIDE = torch.as_tensor(self._config["R_OUTSIDE"])
        R_HALL = torch.as_tensor(self._config["R_HALL"])
        R_WALL = torch.as_tensor(self._config["R_WALL"])

        temp = self.temp
        temp_ = temp + TIME_DELTA / CAP * (
            air * CAP_AIR * (TEMP_AIR - temp) * IS_ROOM
            + ((ADJ | ADJ.T) * (temp[np.newaxis] - temp[np.newaxis].T) / R_WALL).sum(
                dim=-1
            )
            + ADJ_OUTSIDE * (temp_outside - temp) / R_OUTSIDE
            + ADJ_HALL * (temp_hall - temp) / R_HALL
        )

        return temp_

    def _step_time(self, time):
        timestep = torch.round(self._horizon * time)
        return torch.clamp((timestep + 1) / self._horizon, 0, 1)

    def reward_fn(self, state, action, next_state):
        # pylint:disable=unused-argument,missing-docstring
        AIR_MAX = torch.as_tensor(self._config["AIR_MAX"])
        air = torch.as_tensor(action) * AIR_MAX

        temp, _ = self._unpack_state(next_state)

        IS_ROOM = torch.as_tensor(self._config["IS_ROOM"])
        COST_AIR = torch.as_tensor(self._config["COST_AIR"])
        TEMP_LOW = torch.as_tensor(self._config["TEMP_LOW"])
        TEMP_UP = torch.as_tensor(self._config["TEMP_UP"])
        PENALTY = torch.as_tensor(self._config["PENALTY"])

        reward = -(
            IS_ROOM
            * (
                air * COST_AIR
                + ((temp < TEMP_LOW) | (temp > TEMP_UP)) * PENALTY
                + 10.0 * torch.abs((TEMP_UP + TEMP_LOW) / 2.0 - temp)
            )
        ).sum(dim=-1)

        return reward

    def _terminal(self):
        _, time = self._unpack_state(self._state)
        return time.item() >= 1.0

    @staticmethod
    def _unpack_state(state):
        obs = torch.as_tensor(state[..., :-1], dtype=torch.float32)
        time = torch.as_tensor(state[..., -1:], dtype=torch.float32)
        return obs, time

    def render(self, mode="human"):
        pass



"""Trainer and configuration for SOP."""
from raylab.agents import Trainer
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import SOPTorchPolicy


@configure
@OffPolicyMixin.add_options
class SOPTrainer(OffPolicyMixin, Trainer):
    """Single agent trainer for the Streamlined Off-Policy algorithm."""

    _name = "SOP"
    _policy_class = SOPTorchPolicy



# pylint:disable=missing-module-docstring
import gym
import numpy as np
from gym.spaces import Box

from .mixins import IrrelevantRedundantMixin, RNGMixin
from .utils import assert_flat_box_space


class RandomIrrelevant(IrrelevantRedundantMixin, RNGMixin, gym.ObservationWrapper):
    """Add Normal random variables to the environment's observations.

    Args:
        env: Gym environment
        size: Number of random reward-irrelevant variables
        loc: Normal mean
        scale: Normal standard deviation
    """

    def __init__(self, env: gym.Env, size: int, loc: float = 0.0, scale: float = 1.0):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)
        self._size = size
        self._loc = loc
        self._scale = scale

        original = self.observation_space
        low = np.concatenate([original.low, [-np.inf] * self._size])
        high = np.concatenate([original.high, [np.inf] * self._size])
        self.observation_space = Box(
            low=low.astype(original.dtype), high=high.astype(original.dtype)
        )

        self._set_reward_if_possible()
        self._set_termination_if_possible()

    @property
    def added_size(self):
        return self._size

    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        return self.np_random.normal(loc=self._loc, scale=self._scale, size=self._size)



# pylint:disable=missing-module-docstring
from typing import Optional

import gym
import numpy as np
from gym.spaces import Box

from .mixins import IrrelevantRedundantMixin, RNGMixin
from .utils import assert_flat_box_space, check_redundant_size_compat


class LinearRedundant(IrrelevantRedundantMixin, RNGMixin, gym.ObservationWrapper):
    """Adds linear redundant variables to the environment's observations.

    Computes redundant variables as a linear function, with weight matrix
    :math:`W`, of the observation features. Samples :math:`W` upon reset from a
    uniform distribution :math:`Unif_{d\times d}(0, 1)`, where :math:`d` is the
    observation size.

    Args:
        env: Gym environment instance
        size: Number of left-most features from the observation to use in
            computing redundant variables. Defaults to the observation size
    """

    def __init__(self, env: gym.Env, size: Optional[int] = None):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)
        original = self.env.observation_space
        self._size = size = size or original.shape[0]
        check_redundant_size_compat(size, original)

        self._wmat: np.ndarray = None

        low = np.concatenate([original.low, [-np.inf] * size]).astype(original.dtype)
        high = np.concatenate([original.high, [np.inf] * size]).astype(original.dtype)
        self.observation_space = Box(low=low, high=high)

        self._set_reward_if_possible()
        self._set_termination_if_possible()

    @property
    def added_size(self):
        return self.env.observation_space.shape[0]

    def reset(self, **kwargs) -> np.ndarray:
        size = self._size
        self._wmat = self.np_random.uniform(low=0.0, high=1.0, size=(size, size))
        return super().reset(**kwargs)

    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        return self._wmat @ observation[: self._size]



"""TD3 policy class using PyTorch."""
import torch
from nnrl.nn.critic import HardValue
from nnrl.nn.utils import update_polyak
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib.utils import override
from torch.nn.utils import clip_grad_norm_

from raylab.options import configure, option
from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import DeterministicPolicyGradient, FittedQLearning
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options


@configure
@off_policy_options
@option(
    "policy_delay",
    2,
    help="Update policy every this number of calls to `learn_on_batch`",
)
@option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@option("module/type", "TD3")
@option("optimizer/actor", {"type": "Adam", "lr": 1e-3})
@option("optimizer/critics", {"type": "Adam", "lr": 1e-3})
@option("exploration_config/type", "raylab.utils.exploration.GaussianNoise")
@option("exploration_config/noise_stddev", 0.3)
class TD3TorchPolicy(OffPolicyMixin, TorchPolicy):
    """TD3 policy in Pytorch for RLlib."""

    dist_class = WrapDeterministicPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._make_actor_loss()
        self._make_critic_loss()
        self._grad_step = 0
        self._info = {}

        self.build_replay_buffer()

    def _make_actor_loss(self):
        self.loss_actor = DeterministicPolicyGradient(
            self.module.actor,
            self.module.critics,
        )

    def _make_critic_loss(self):
        target_value = HardValue(self.module.target_actor, self.module.target_critics)
        self.loss_critic = FittedQLearning(self.module.critics, target_value)
        self.loss_critic.gamma = self.config["gamma"]

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        optimizers["actor"] = build_optimizer(self.module.actor, config["actor"])
        optimizers["critics"] = build_optimizer(self.module.critics, config["critics"])
        return optimizers

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict) -> dict:
        self._grad_step += 1
        self._info["grad_steps"] = self._grad_step

        self._info.update(self._update_critic(batch))
        if self._grad_step % self.config["policy_delay"] == 0:
            self._info.update(self._update_policy(batch))

        return self._info.copy()

    def _update_critic(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("critics"):
            loss, info = self.loss_critic(batch)
            loss.backward()
            info.update(self.extra_grad_info("critics"))

        main, target = self.module.critics, self.module.target_critics
        update_polyak(main, target, self.config["polyak"])
        return info

    def _update_policy(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("actor"):
            loss, info = self.loss_actor(batch)
            loss.backward()
            info.update(self.extra_grad_info("actor"))

        main, target = self.module.actor, self.module.target_actor
        update_polyak(main, target, self.config["polyak"])
        return info

    @torch.no_grad()
    def extra_grad_info(self, component: str) -> dict:
        """Return gradient statistics for the given component."""
        params = getattr(self.module, component).parameters()
        clip = float("inf")
        return {f"grad_norm({component})": clip_grad_norm_(params, clip).item()}

    @override(TorchPolicy)
    def get_weights(self) -> dict:
        weights = super().get_weights()
        weights["grad_step"] = self._grad_step
        return weights

    @override(TorchPolicy)
    def set_weights(self, weights: dict):
        self._grad_step = weights["grad_step"]
        super().set_weights(weights)



"""Registry of agents as trainables for Tune."""
from ray.tune.registry import get_trainable_cls


def _import_naf():
    from raylab.agents.naf import NAFTrainer

    return NAFTrainer


def _import_svg_inf():
    from raylab.agents.svg.inf import SVGInfTrainer

    return SVGInfTrainer


def _import_svg_one():
    from raylab.agents.svg.one import SVGOneTrainer

    return SVGOneTrainer


def _import_sac():
    from raylab.agents.sac import SACTrainer

    return SACTrainer


def _import_sop():
    from raylab.agents.sop import SOPTrainer

    return SOPTrainer


def _import_trpo():
    from raylab.agents.trpo import TRPOTrainer

    return TRPOTrainer


def _import_soft_svg():
    from raylab.agents.svg.soft import SoftSVGTrainer

    return SoftSVGTrainer


def _import_acktr():
    from raylab.agents.acktr import ACKTRTrainer

    return ACKTRTrainer


def _import_mbpo():
    from raylab.agents.mbpo import MBPOTrainer

    return MBPOTrainer


def _import_mage():
    from raylab.agents.mage import MAGETrainer

    return MAGETrainer


def _import_td3():
    from raylab.agents.td3 import TD3Trainer

    return TD3Trainer


AGENTS = {
    "NAF": _import_naf,
    "SVG(inf)": _import_svg_inf,
    "SVG(1)": _import_svg_one,
    "SoftAC": _import_sac,
    "SOP": _import_sop,
    "TRPO": _import_trpo,
    "SoftSVG": _import_soft_svg,
    "ACKTR": _import_acktr,
    "MBPO": _import_mbpo,
    "MAGE": _import_mage,
    "raylab/TD3": _import_td3,
}


def get_agent_cls(agent_name):
    """Retrieve agent class from global registry.

    The user must have called `raylab.register_all_agents()` beforehand to
    have access to Raylab's agents.
    """
    return get_trainable_cls(agent_name)



"""Trainer and configuration for SVG(inf)."""
from ray.rllib.utils import override

from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.trainer import Trainer
from raylab.options import configure, option

from .policy import SVGInfTorchPolicy


@configure
@option("rollout_fragment_length", default=1, override=True)
@option("batch_mode", "complete_episodes", override=True)
@option("num_workers", default=0, override=True)
@option("evaluation_config/explore", True)
class SVGInfTrainer(Trainer):
    """Single agent trainer for SVG(inf)."""

    _name = "SVG(inf)"
    _policy_class = SVGInfTorchPolicy

    @override(Trainer)
    def validate_config(self, config: dict):
        """Assert configuration values are valid."""
        super().validate_config(config)
        assert config["num_workers"] == 0, "No point in using additional workers."
        assert (
            config["rollout_fragment_length"] >= 1
        ), "At least one sample must be collected."
        assert (
            config["batch_mode"] == "complete_episodes"
        ), "SVG(inf) uses full rollouts"

    @override(Trainer)
    def after_init(self):
        super().after_init()
        set_policy_with_env_fn(self.workers, fn_type="reward")



"""Utilities for default trainer configurations and descriptions."""
import copy
import inspect
import textwrap
from collections import namedtuple
from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Set, Union

import tree
from dataclasses_json.core import Json
from ray.rllib.agents import with_common_config as with_rllib_config
from ray.rllib.agents.trainer import COMMON_CONFIG, Trainer
from ray.rllib.utils import deep_update, merge_dicts

__all__ = [
    "COMMON_INFO",
    "Config",
    "Info",
    "Json",
    "Option",
    "RaylabOptions",
    "recursive_check_info",
    "with_rllib_info",
]

Config = Dict[str, Union[Json, "Config"]]
Info = Dict[str, Union[str, "Info"]]


# ==============================================================================
# Programatic config setting
# ==============================================================================
def configure(cls: type) -> type:
    """Decorator for finishing the configuration setup for a class.

    Should be called after all :func:`config` decorators have been applied,
    i.e., as the top-most decorator.
    """
    cls.options = cls.options.copy_and_set_queued_options()
    return cls


def option(
    key: str,
    default: Json = None,
    *,
    help: Optional[str] = None,
    override: bool = False,
    allow_unknown_subkeys: bool = False,
    override_all_if_type_changes: bool = False,
    separator: str = "/",
) -> Callable[[type], type]:
    """Returns a decorator for adding/overriding a class config.

    If `key` ends in a separator and `default` is None, treats the option as a
    nested dict of options and sets the default to an empty dictionary.

    Args:
        key: Name of the config paremeter which the user can tune
        default: Default Jsonable value to set for the parameter
        info: Parameter help text explaining what the parameter does
        override: Whether to override an existing parameter
        allow_unknown_subkeys: Whether to allow new keys for dict parameters.
            This is only at the top level
        override_all_if_type_changes: Whether to override the entire value
            (dict) iff the 'type' key in this value dict changes. This is only
            at the top level
        separator: String token separating nested keys

    Raises:
        RuntimeError: If attempting to set an existing parameter with `override`
            set to `False`.
    """
    # pylint:disable=too-many-arguments,redefined-builtin
    def _queue(cls):
        cls.options.add_option_to_queue(
            key=key,
            default=default,
            info=help,
            override=override,
            allow_unknown_subkeys=allow_unknown_subkeys,
            override_all_if_type_changes=override_all_if_type_changes,
            separator=separator,
        )
        return cls

    return _queue


# ==============================================================================
# Configuration objects
# ==============================================================================
Option = namedtuple("Option", "key parameters")


@dataclass
class RaylabOptions:
    """Configuration object for Raylab objects.

    Attributes:
        defaults: Default configurations
        infos: Help msgs for each configuration
    """

    defaults: Config = field(default_factory=dict)
    infos: Info = field(default_factory=dict)
    allow_unknown_subkeys: Set[str] = field(default_factory=set)
    override_all_if_type_changes: Set[str] = field(default_factory=set)
    dict_info_key: str = field(default="__help__", init=False, repr=False)
    _options_to_set: List[Option] = field(default_factory=list, init=False, repr=False)

    @property
    def all_options_set(self) -> bool:
        """Whether all queued options have been set and returned as a new config."""
        return not bool(self._options_to_set)

    def add_option_to_queue(
        self,
        key: str,
        default: Json = None,
        info: Optional[str] = None,
        override: bool = False,
        allow_unknown_subkeys: bool = False,
        override_all_if_type_changes: bool = False,
        separator: str = "/",
    ):
        """Queues an option with the given parameters to be set for a new config.

        Args:
            key: Name of the config option which the user can tune
            default: Default Jsonable value to set for the option
            info: Help text explaining what the option does
            override: Whether to override an existing option default
            allow_unknown_subkeys: Whether to allow new keys for dict options.
                This is only at the top level
            override_all_if_type_changes: Whether to override the entire value
                (dict) iff the 'type' key in this value dict changes. This is
                only at the top level
            separator: String token separating nested keys

        Raises:
            RuntimeError: If attempting to enable `allow_unknown_subkeys` or
                `override_all_if_type_changes` options for non-toplevel keys
        """
        # pylint:disable=too-many-arguments
        should_be_toplevel = allow_unknown_subkeys or override_all_if_type_changes
        if should_be_toplevel and separator in key.rstrip(separator):
            raise RuntimeError(
                "Cannot use 'allow_unknown_subkeys' or 'override_all_if_type_changes'"
                f" for non-toplevel key: '{key}'"
            )

        self._options_to_set.append(
            Option(
                key=key,
                parameters=dict(
                    default=default,
                    info=info,
                    override=override,
                    allow_unknown_subkeys=allow_unknown_subkeys,
                    override_all_if_type_changes=override_all_if_type_changes,
                    separator=separator,
                ),
            )
        )

    def copy_and_set_queued_options(self) -> "RaylabOptions":
        """Create a new RaylabOptions and set all queued options.

        Returns:
            A new RaylabOptions instance with this instance's defaults and all
            options queued via :meth:`add_option_to_queue`.

        Raises:
            See :meth:`set_option`.
        """
        new = type(self)(
            defaults=copy.deepcopy(self.defaults),
            infos=copy.deepcopy(self.infos),
            allow_unknown_subkeys=copy.deepcopy(self.allow_unknown_subkeys),
            override_all_if_type_changes=copy.deepcopy(
                self.override_all_if_type_changes
            ),
        )

        to_set = self._options_to_set
        self._options_to_set = []
        for opt in sorted(to_set, key=lambda x: x.key):
            new.set_option(key=opt.key, **opt.parameters)

        return new

    def merge_defaults_with(self, config: Config) -> Config:
        """Deep merge the given config with the defaults."""
        defaults = copy.deepcopy(self.defaults)
        new = deep_update(
            defaults,
            config,
            new_keys_allowed=False,
            allow_new_subkey_list=self.allow_unknown_subkeys,
            override_all_if_type_changes=self.override_all_if_type_changes,
        )
        return new

    def set_option(
        self,
        key: str,
        default: Json = None,
        info: Optional[str] = None,
        override: bool = False,
        allow_unknown_subkeys: bool = False,
        override_all_if_type_changes: bool = False,
        separator: str = "/",
    ):
        """Set an option in-place for this config.

        If `key` ends in a separator and `default` is None, treats the option as
        a nested dict of options and sets the default to an empty dictionary
        (unless overriding an existing option).

        Raises:
            ValueError: If attempting to set an existing option with `override`
                set to `False`.
            ValueError: If attempting to override an existing option with its
                same default value.
        """
        # pylint:disable=too-many-arguments
        if key.endswith(separator):
            if not override and not isinstance(default, (dict, type(None))):
                raise ValueError(
                    f"Key '{key}' ends in a separator by default is neither None or"
                    f" a dictionary: {default}"
                )
            key = key.rstrip(separator)
            default = {} if not override and default is None else default

        key_seq = key.split(separator)

        if allow_unknown_subkeys and not override:
            self.allow_unknown_subkeys.add(key)
        if override_all_if_type_changes and not override:
            self.override_all_if_type_changes.add(key)

        config_, info_ = self.defaults, self.infos
        for key_ in key_seq[:-1]:
            config_ = config_.setdefault(key_, {})
            info_ = info_.setdefault(key_, {})
        key_ = key_seq[-1]

        if key_ in config_ and not override:
            raise ValueError(
                f"Attempted to override config key '{key_}' but override=False."
            )
        if key_ in config_ and default == config_[key_]:
            raise ValueError(
                f"Attempted to override config key {key} with the same value: {default}"
            )
        if key_ not in config_ and override:
            raise ValueError(f"Attempted to override inexistent config key '{key}'")
        config_[key_] = default

        if info is not None:
            help_txt = inspect.cleandoc(info)
            if isinstance(config_[key_], dict):
                info_[key_] = {self.dict_info_key: help_txt}
            else:
                info_[key_] = help_txt

    # ==========================================================================
    # Help with options
    # ==========================================================================
    def help(self, key: Optional[str] = None, separator: str = "/") -> str:
        """Returns a help text for options.

        Args:
            key: String name of the option to be described. If None, lists all
                toplevel options descriptions
            separator: String token separating nested option names

        Raises:
            UnknownOptionError: If failing to find a specific given `key`
        """
        if key is not None:
            return self.find_config_info(key, separator)

        toplevel_keys = set(self.infos.keys())

        msgs = [
            self.parse_info(self.defaults, self.infos, k) for k in sorted(toplevel_keys)
        ]
        return "\n".join(msgs)

    def find_config_info(self, key: str, separator: str) -> str:
        """Find option description in info dict.

        Args:
            key: Hierarchy of nested parameter keys leading to the desired key
            separator: Text token separating nested info keys

        Returns:
            The parameter's help text.

        Raises:
            UnknownOptionError: If the search fails at any point in the key
                sequence
        """
        key_seq = key.split(separator)
        config = self.defaults
        info = self.infos

        def check_help(k, i, seq):
            if k not in i:
                k_str = separator.join(seq)
                raise UnknownOptionError(k_str)

        for idx, key_ in enumerate(key_seq[:-1]):
            check_help(key_, info, key_seq[: idx + 1])
            config = config[key_]
            info = info[key_]
        key_ = key_seq[-1]
        check_help(key_, info, key_seq)

        # Info documents a single value
        if isinstance(info[key_], str):
            return self.parse_info(config, info, key_)

        # Info documents a nested config
        config, info = config[key_], info[key_]
        msgs = info.get(self.dict_info_key, "")
        msgs += [
            self.parse_info(config, info, k)
            for k in info.keys()
            if k != self.dict_info_key
        ]
        return "\n".join(msgs)

    def parse_info(
        self, config: dict, info: Dict[str, Union[str, dict]], key: str
    ) -> str:
        """Returns the string form of the option info.

        Uses a hardcoded maximum default representation length of 40.
        """
        default = repr(config[key])
        if len(default) > 40:
            default = repr(type(config[key]))

        help_ = info[key]
        if isinstance(help_, dict):
            help_ = help_[self.dict_info_key]

        msg = f"{key}: {default}\n"
        msg = msg + textwrap.indent(f"{help_}", prefix=" " * 4)
        return msg


@dataclass
class TrainerOptions(RaylabOptions):
    """Configuration object for Raylab trainers."""

    defaults: Config = field(default_factory=lambda: with_rllib_config({}))
    infos: Info = field(default_factory=lambda: with_rllib_info({}))
    # pylint:disable=protected-access
    allow_unknown_subkeys: Set[str] = field(
        default_factory=lambda: set(Trainer._allow_unknown_subkeys)
    )
    override_all_if_type_changes: Set[str] = field(
        default_factory=lambda: set(Trainer._override_all_subkeys_if_type_changes)
    )
    # pylint:enable=protected-access
    _rllib_keys: Set[str] = field(
        default_factory=lambda: set(COMMON_CONFIG.keys()), init=False, repr=False
    )

    @property
    def rllib_defaults(self) -> Config:
        """Default configurations for RLlib trainers."""
        return {k: v for k, v in self.defaults.items() if k in self._rllib_keys}

    def rllib_subconfig(self, config: dict) -> dict:
        """Get the rllib subconfig from `config`."""
        return {k: v for k, v in config.items() if k in self._rllib_keys}

    def help(
        self, key: Optional[str] = None, separator: str = "/", with_rllib: bool = False
    ) -> str:
        """Returns a help text for options.

        Args:
            key: String name of the option to be described. If None, lists all
                toplevel options descriptions
            separator: String token separating nested option names
            with_rllib: Whether to include RLlib's common config descriptions if
                listing all options descriptions

        Raises:
            UnknownOptionError: If failing to find a specific given `key`
        """
        # pylint:disable=arguments-differ
        if key is not None:
            return self.find_config_info(key, separator)

        toplevel_keys = set(self.infos.keys())
        if not with_rllib:
            toplevel_keys.difference_update(set(self._rllib_keys))

        msgs = [
            self.parse_info(self.defaults, self.infos, k) for k in sorted(toplevel_keys)
        ]
        return "\n".join(msgs)


# ================================================================================
# RLlib Help
# ================================================================================


COMMON_INFO = {
    # === Settings for Rollout Worker processes ===
    "num_workers": """\
    Number of rollout worker actors to create for parallel sampling. Setting
    this to 0 will force rollouts to be done in the trainer actor.""",
    "num_envs_per_worker": """\
    Number of environments to evaluate vectorwise per worker. This enables
    model inference batching, which can improve performance for inference
    bottlenecked workloads.""",
    "create_env_on_driver": """\
    When `num_workers` > 0, the driver (local_worker; worker-idx=0) does not
    need an environment. This is because it doesn't have to sample (done by
    remote_workers; worker_indices > 0) nor evaluate (done by evaluation
    workers; see below).""",
    "rollout_fragment_length": """\
    Divide episodes into fragments of this many steps each during rollouts.
    Sample batches of this size are collected from rollout workers and
    combined into a larger batch of `train_batch_size` for learning.

    For example, given rollout_fragment_length=100 and train_batch_size=1000:
      1. RLlib collects 10 fragments of 100 steps each from rollout workers.
      2. These fragments are concatenated and we perform an epoch of SGD.

    When using multiple envs per worker, the fragment size is multiplied by
    `num_envs_per_worker`. This is since we are collecting steps from
    multiple envs in parallel. For example, if num_envs_per_worker=5, then
    rollout workers will return experiences in chunks of 5*100 = 500 steps.

    The dataflow here can vary per algorithm. For example, PPO further
    divides the train batch into minibatches for multi-epoch SGD.""",
    "batch_mode": """\
    Whether to rollout "complete_episodes" or "truncate_episodes" to
    `rollout_fragment_length` length unrolls. Episode truncation guarantees
    evenly sized batches, but increases variance as the reward-to-go will
    need to be estimated at truncation boundaries.""",
    # === Settings for the Trainer process ===
    "num_gpus": """\
    Number of GPUs to allocate to the trainer process. Note that not all
    algorithms can take advantage of trainer GPUs. This can be fractional
    (e.g., 0.3 GPUs).""",
    "train_batch_size": """\
    Training batch size, if applicable. Should be >= rollout_fragment_length.
    Samples batches will be concatenated together to a batch of this size,
    which is then passed to SGD.""",
    "model": """\
    Arguments to pass to the policy model. See models/catalog.py for a full
    list of the available model options.""",
    "optimizer": """\
    Arguments to pass to the policy optimizer. These vary by optimizer.""",
    # === Environment Settings ===
    "gamma": """\
    Discount factor of the MDP.""",
    "horizon": """\
    Number of steps after which the episode is forced to terminate. Defaults
    to `env.spec.max_episode_steps` (if present) for Gym envs.""",
    "soft_horizon": """\
    Calculate rewards but don't reset the environment when the horizon is
    hit. This allows value estimation and RNN state to span across logical
    episodes denoted by horizon. This only has an effect if horizon != inf.""",
    "no_done_at_end": """\
    Don't set 'done' at the end of the episode. Note that you still need to
    set this if soft_horizon=True, unless your env is actually running
    forever without returning done=True.""",
    "env_config": """\
    Arguments to pass to the env creator.""",
    "env_task_fn": """\
    A callable taking the last train results, the base env and the env
    context as args and returning a new task to set the env to.
    The env must be a `TaskSettableEnv` sub-class for this to work.
    See `examples/curriculum_learning.py` for an example.
    """,
    "render_env": """\
    If True, try to render the environment on the local worker or on worker
    1 (if num_workers > 0). For vectorized envs, this usually means that only
    the first sub-environment will be rendered.
    In order for this to work, your env will have to implement the
    `render()` method which either:
    a) handles window generation and rendering itself (returning True) or
    b) returns a numpy uint8 image of shape [height x width x 3 (RGB)].
    """,
    "record_env": """\
    If True, stores videos in this relative directory inside the default
    output dir (~/ray_results/...). Alternatively, you can specify an
    absolute path (str), in which the env recordings should be
    stored instead.
    Set to False for not recording anything.
    Note: This setting replaces the deprecated `monitor` key.
    """,
    "env": """\
    Environment name can also be passed via config.""",
    "observation_space": """\
    The observation- and action spaces for the Policies of this Trainer.
    Use None for automatically inferring these from the given env.
    """,
    "action_space": """\
    The observation- and action spaces for the Policies of this Trainer.
    Use None for automatically inferring these from the given env.
    """,
    "normalize_actions": """\
    Unsquash actions to the upper and lower bounds of env's action space""",
    "clip_rewards": """\
    Whether to clip rewards prior to experience postprocessing. Setting to
    None means clip for Atari only.""",
    "clip_actions": """\
    Whether to np.clip() actions to the action space low/high range spec.""",
    "preprocessor_pref": """\
    Whether to use rllib or deepmind preprocessors by default""",
    "lr": """\
    The default learning rate. Not used by Raylab""",
    # === Debug Settings ===
    "simple_optimizer": """\
    # Uses the sync samples optimizer instead of the multi-gpu one. This is
    # usually slower, but you might want to try it if you run into issues with
    # the default optimizer.
    # This will be set automatically from now on.
    """,
    "monitor": """\
    Whether to write episode stats and videos to the agent log dir. This is
    typically located in ~/ray_results.""",
    "log_level": """\
    Set the ray.rllib.* log level for the agent process and its workers.
    Should be one of DEBUG, INFO, WARN, or ERROR. The DEBUG level will also
    periodically print out summaries of relevant internal dataflow (this is
    also printed out once at startup at the INFO level). When using the
    `rllib train` command, you can also use the `-v` and `-vv` flags as
    shorthand for INFO and DEBUG.""",
    "callbacks": """\
    Callbacks that will be run during various phases of training. See the
    `DefaultCallbacks` class and `examples/custom_metrics_and_callbacks.py`
    for more usage information.""",
    "ignore_worker_failures": """\
    Whether to attempt to continue training if a worker crashes. The number
    of currently healthy workers is reported as the "num_healthy_workers"
    metric.""",
    "log_sys_usage": """\
    Log system resource metrics to results. This requires `psutil` to be
    installed for sys stats, and `gputil` for GPU metrics.""",
    "fake_sampler": """\
    Use fake (infinite speed) sampler. For testing only.""",
    # === Deep Learning Framework Settings ===
    "framework": """\
    tf: TensorFlow
    tfe: TensorFlow eager
    torch: PyTorch""",
    "eager_tracing": """\
    Enable tracing in eager mode. This greatly improves performance, but
    makes it slightly harder to debug since Python code won't be evaluated
    after the initial eager pass. Only possible if framework=tfe.""",
    # === Exploration Settings ===
    "explore": """\
    Default exploration behavior, iff `explore`=None is passed into
    compute_action(s).
    Set to False for no exploration behavior (e.g., for evaluation).""",
    "exploration_config": {
        RaylabOptions.dict_info_key: """\
        Provide a dict specifying the Exploration object's config.
        """,
        "type": """\
        The Exploration class to use. In the simplest case, this is the name
        (str) of any class present in the `rllib.utils.exploration` package.
        You can also provide the python class directly or the full location
        of your class (e.g. "ray.rllib.utils.exploration.epsilon_greedy.
        EpsilonGreedy").
        """,
    },
    # === Evaluation Settings ===
    "evaluation_interval": """\
    Evaluate with every `evaluation_interval` training iterations.
    The evaluation stats will be reported under the "evaluation" metric key.
    Note that evaluation is currently not parallelized, and that for Ape-X
    metrics are already only reported for the lowest epsilon workers.""",
    "evaluation_num_episodes": """\
    Number of episodes to run per evaluation period. If using multiple
    evaluation workers, we will run at least this many episodes total.""",
    "evaluation_parallel_to_training": """\
    Whether to run evaluation in parallel to a Trainer.train() call
    using threading. Default=False.
    E.g. evaluation_interval=2 -> For every other training iteration,
    the Trainer.train() and Trainer.evaluate() calls run in parallel.
    Note: This is experimental. Possible pitfalls could be race conditions
    for weight synching at the beginning of the evaluation loop.
    """,
    "in_evaluation": """\
    Internal flag that is set to True for evaluation workers.""",
    "evaluation_config": """\
    Typical usage is to pass extra args to evaluation env creator
    and to disable exploration by computing deterministic actions.
    IMPORTANT NOTE: Policy gradient algorithms are able to find the optimal
    policy, even if this is a stochastic one. Setting "explore=False" here
    will result in the evaluation workers not using this optimal policy!

    Example: overriding env_config, exploration, etc:
    "env_config": {...},
    "explore": False""",
    "evaluation_num_workers": """\
    Number of parallel workers to use for evaluation. Note that this is set
    to zero by default, which means evaluation will be run in the trainer
    process. If you increase this, it will increase the Ray resource usage
    of the trainer since evaluation workers are created separately from
    rollout workers.""",
    "custom_eval_function": """\
    Customize the evaluation method. This must be a function of signature
    (trainer: Trainer, eval_workers: WorkerSet) -> metrics: dict. See the
    Trainer._evaluate() method to see the default implementation. The
    trainer guarantees all eval workers have the latest policy state before
    this function is called.""",
    # === Advanced Rollout Settings ===
    "sample_async": """\
    Use a background thread for sampling (slightly off-policy, usually not
    advisable to turn on unless your env specifically requires it).""",
    "sample_collector": """\
    The SampleCollector class to be used to collect and retrieve
    environment-, model-, and sampler data. Override the SampleCollector base
    class to implement your own collection/buffering/retrieval logic.
    """,
    "observation_filter": """\
    Element-wise observation filter, either "NoFilter" or "MeanStdFilter".""",
    "synchronize_filters": """\
    Whether to synchronize the statistics of remote filters.""",
    "tf_session_args": {
        RaylabOptions.dict_info_key: """\
        Configures TF for single-process operation by default.
        """,
        "intra_op_parallelism_threads": """\
        note: overriden by `local_tf_session_args`
        """,
        "inter_op_parallelism_threads": "",
        "gpu_options": {"allow_growth": ""},
        "log_device_placement": "",
        "device_count": {"CPU": ""},
        "allow_soft_placement": """\
        required by PPO multi-gpu
        """,
    },
    "local_tf_session_args": {
        RaylabOptions.dict_info_key: """\
        Override the following tf session args on the local worker

        Allow a higher level of parallelism by default, but not unlimited
        since that can cause crashes with many concurrent drivers.
        """,
        "intra_op_parallelism_threads": "",
        "inter_op_parallelism_threads": "",
    },
    "compress_observations": """\
    Whether to LZ4 compress individual observations""",
    "collect_metrics_timeout": """\
    Wait for metric batches for at most this many seconds. Those that
    have not returned in time will be collected in the next train iteration.""",
    "metrics_smoothing_episodes": """\
    Smooth metrics over this many episodes.""",
    "remote_worker_envs": """\
    If using num_envs_per_worker > 1, whether to create those new envs in
    remote processes instead of in the same worker. This adds overheads, but
    can make sense if your envs can take much time to step / reset
    (e.g., for StarCraft). Use this cautiously; overheads are significant.""",
    "remote_env_batch_wait_ms": """\
    Timeout that remote workers are waiting when polling environments.
    0 (continue when at least one env is ready) is a reasonable default,
    but optimal value could be obtained by measuring your environment
    step / reset and model inference perf.""",
    "min_iter_time_s": """\
    Minimum time per train iteration (frequency of metrics reporting).""",
    "timesteps_per_iteration": """\
    Minimum env steps to optimize for per train call. This value does
    not affect learning, only the length of train iterations.""",
    "seed": """\
    This argument, in conjunction with worker_index, sets the random seed of
    each worker, so that identically configured trials will have identical
    results. This makes experiments reproducible.""",
    "extra_python_environs_for_driver": """\
    Any extra python env vars to set in the trainer process, e.g.,
    {"OMP_NUM_THREADS": "16"}""",
    "extra_python_environs_for_worker": """\
    The extra python environments need to set for worker processes.""",
    # === Advanced Resource Settings ===
    "_fake_gpus": """\
    Set to True for debugging (multi-)?GPU funcitonality on a CPU machine.
    GPU towers will be simulated by graphs located on CPUs in this case.
    Use `num_gpus` to test for different numbers of fake GPUs.
    """,
    "num_cpus_per_worker": """\
    Number of CPUs to allocate per worker.""",
    "num_gpus_per_worker": """\
    Number of GPUs to allocate per worker. This can be fractional. This is
    usually needed only if your env itself requires a GPU (i.e., it is a
    GPU-intensive video game), or model inference is unusually expensive.""",
    "custom_resources_per_worker": """\
    Any custom Ray resources to allocate per worker.""",
    "num_cpus_for_driver": """\
    Number of CPUs to allocate for the trainer. Note: this only takes effect
    when running in Tune. Otherwise, the trainer runs in the main program.""",
    "placement_strategy": """\
    The strategy for the placement group factory returned by
    `Trainer.default_resource_request()`. A PlacementGroup defines, which
    devices (resources) should always be co-located on the same node.
    For example, a Trainer with 2 rollout workers, running with
    num_gpus=1 will request a placement group with the bundles:
    [{"gpu": 1, "cpu": 1}, {"cpu": 1}, {"cpu": 1}], where the first bundle is
    for the driver and the other 2 bundles are for the two workers.
    These bundles can now be "placed" on the same or different
    nodes depending on the value of `placement_strategy`:
    "PACK": Packs bundles into as few nodes as possible.
    "SPREAD": Places bundles across distinct nodes as even as possible.
    "STRICT_PACK": Packs bundles into one node. The group is not allowed
      to span multiple nodes.
    "STRICT_SPREAD": Packs bundles across distinct nodes.
    """,
    # === Offline Datasets ===
    "input": """\
    Specify how to generate experiences:
     - "sampler": generate experiences via online simulation (default)
     - a local directory or file glob expression (e.g., "/tmp/*.json")
     - a list of individual file paths/URIs (e.g., ["/tmp/1.json",
       "s3://bucket/2.json"])
     - a dict with string keys and sampling probabilities as values (e.g.,
       {"sampler": 0.4, "/tmp/*.json": 0.4, "s3://bucket/expert.json": 0.2}).
     - a function that returns a rllib.offline.InputReader""",
    "input_config": """\
    Arguments accessible from the IOContext for configuring custom input
    """,
    "actions_in_input_normalized": """\
    True, if the actions in a given offline "input" are already normalized
    (between -1.0 and 1.0). This is usually the case when the offline
    file has been generated by another RLlib algorithm (e.g. PPO or SAC),
    while "normalize_actions" was set to True.
    """,
    "input_evaluation": """\
    Specify how to evaluate the current policy. This only has an effect when
    reading offline experiences. Available options:
     - "wis": the weighted step-wise importance sampling estimator.
     - "is": the step-wise importance sampling estimator.
     - "simulation": run the environment in the background, but use
       this data for evaluation only and not for learning.""",
    "postprocess_inputs": """\
    Whether to run postprocess_trajectory() on the trajectory fragments from
    offline inputs. Note that postprocessing will be done using the *current*
    policy, not the *behavior* policy, which is typically undesirable for
    on-policy algorithms.""",
    "shuffle_buffer_size": """\
    If positive, input batches will be shuffled via a sliding window buffer
    of this number of batches. Use this if the input data is not in random
    enough order. Input is delayed until the shuffle buffer is filled.""",
    "output": """\
    Specify where experiences should be saved:
     - None: don't save any experiences
     - "logdir" to save to the agent log dir
     - a path/URI to save to a custom output directory (e.g., "s3://bucket/")
     - a function that returns a rllib.offline.OutputWriter""",
    "output_compress_columns": """\
    What sample batch columns to LZ4 compress in the output data.""",
    "output_max_file_size": """\
    Max output file size before rolling over to a new file.""",
    # === Settings for Multi-Agent Environments ===
    "multiagent": {
        RaylabOptions.dict_info_key: """\
        === Settings for Multi-Agent Environments ===
        """,
        "policies": """\
        Map of type MultiAgentPolicyConfigDict from policy ids to tuples
        of (policy_cls, obs_space, act_space, config). This defines the
        observation and action spaces of the policies and any extra config.
        """,
        "policy_mapping_fn": """\
        Function mapping agent ids to policy ids.
        """,
        "policies_to_train": """\
        Optional list of policies to train, or None for all policies.
        """,
        "observation_fn": """\
        Optional function that can be used to enhance the local agent
        observations to include more state.
        See rllib/evaluation/observation_function.py for more info.
        """,
        "replay_mode": """\
        When replay_mode=lockstep, RLlib will replay all the agent
        transitions at a particular timestep together in a batch. This allows
        the policy to implement differentiable shared computations between
        agents it controls at that timestep. When replay_mode=independent,
        transitions are replayed independently per policy.
        """,
    },
    # === Logger ===
    "logger_config": """\
    Define logger-specific configuration to be used inside Logger
    Default value None allows overwriting with nested dicts
    """,
}


def with_rllib_info(info: Info) -> Info:
    """Merge info with RLlib's common parameters' info."""
    info = merge_dicts(COMMON_INFO, info)
    info = tree.map_structure(
        (lambda x: inspect.cleandoc(x) if isinstance(x, str) else x), info
    )
    return info


# ==============================================================================
# Debuggin utilities
# ==============================================================================
class UnknownOptionError(Exception):
    """Exception raised for querying the description of an unknown option key.

    Args:
        key: Option key (possibly nested) that is not in the agent's options.
    """

    def __init__(self, key: str):
        super().__init__(f"Key {key} could not be found among option descriptions.")


class MissingConfigInfoError(Exception):
    """Exception raised for undocumented config parameter.

    Args:
        key: Name of config parameter which is missing in info dict
    """

    def __init__(self, key: str):
        super().__init__(f"Info does not document {key} config parameter(s).")


class ExtraConfigInfoError(Exception):
    """Exception raised for documented inexistent config parameter.

    Args:
        key: Name of info key which does not exist in config
    """

    def __init__(self, key: str):
        super().__init__(f"Info key(s) {key} does no exist in config dict.")


def recursive_check_info(
    config: Config,
    info: Info,
    allow_new_subkey_list: Optional[List[str]] = None,
    prefix: str = "",
):
    """Recursively check if all keys in config are documented in info.

    Args:
        config: Configuration dictionary with default values
        info: Help dictionary for parameters
        allow_new_subkey_list: List of sub-dictionary keys with arbitrary items.
            Skips recursive check of these keys.
        prefix: String for keeping track of parent config keys. Used when
            raising errors.

    Raises:
        MissingConfigInfoError: If a parameter is not documented
        ExtraConfigInfoError: If info documents an inexistent parameter
    """
    allow_new_subkey_list = allow_new_subkey_list or []

    config_keys = set(config.keys())
    info_keys = set(info.keys())

    missing = config_keys.difference(info_keys)
    if missing:
        raise MissingConfigInfoError(prefix + f"{missing}")

    extra = info_keys.difference(config_keys)
    if extra:
        raise ExtraConfigInfoError(prefix + f"{extra}")

    for key in config.keys():
        config_, info_ = config[key], info[key]
        if isinstance(config_, dict) and isinstance(info_, dict):
            if RaylabOptions.dict_info_key not in info_:
                raise MissingConfigInfoError(prefix + key)
            if key in allow_new_subkey_list:
                continue
            recursive_check_info(config_, info_, prefix=key + "/")



"""STREAMLINED OFF-POLICY LEARNING

From the paper:
TOWARDS SIMPLICITY IN DEEP REINFORCEMENT LEARNING: STREAMLINED OFF-POLICY LEARNING
https://arxiv.org/abs/1910.02208.
"""
from .policy import SOPTorchPolicy
from .trainer import SOPTrainer



# pylint:disable=missing-module-docstring
import logging
import time
from abc import ABCMeta
from typing import Callable, Iterable, Optional, Type

from ray.exceptions import RayError
from ray.rllib import Policy
from ray.rllib.agents import Trainer as RLlibTrainer
from ray.rllib.agents.trainer import MAX_WORKER_FAILURE_RETRIES
from ray.rllib.agents.trainer_template import default_execution_plan
from ray.rllib.env.env_context import EnvContext
from ray.rllib.evaluation.worker_set import WorkerSet
from ray.rllib.utils import override as overrides
from ray.rllib.utils.typing import (
    EnvType,
    PartialTrainerConfigDict,
    ResultDict,
    TrainerConfigDict,
)
from ray.tune.resources import Resources
from ray.tune.trainable import Trainable

from raylab.options import TrainerOptions, configure, option

logger = logging.getLogger(__name__)


# ==============================================================================
# Streamlined Trainer
# ==============================================================================
@configure
@option(
    "policy/",
    allow_unknown_subkeys=True,
    help="""Sub-configurations for the policy class.""",
)
@option("framework", default="torch", override=True)
@option(
    "simple_optimizer",
    # RLlib attempts to use Tensorflow if this is true
    default=True,
    override=True,
)
class Trainer(RLlibTrainer, metaclass=ABCMeta):
    """Base class for raylab trainers."""

    config: TrainerConfigDict
    raw_user_config: PartialTrainerConfigDict
    env_creator: Callable[[EnvContext], EnvType]
    global_vars: dict
    workers: WorkerSet
    evaluation_workers: Optional[WorkerSet]
    train_exec_impl: Iterable[ResultDict]
    options: TrainerOptions = TrainerOptions()
    _name: str
    _policy_class: Type[Policy]
    _env_id: str
    _true_config: TrainerConfigDict

    def setup(self, config: PartialTrainerConfigDict):
        if self._env_id:
            config["env"] = self._env_id
        self._true_config = self.options.merge_defaults_with(config)
        super().setup(self.options.rllib_subconfig(self._true_config))

    @property
    def _default_config(self) -> TrainerConfigDict:
        return self.options.rllib_defaults

    def _init(
        self,
        config: PartialTrainerConfigDict,
        env_creator: Callable[[EnvContext], EnvType],
    ):
        self.config = config = self.restore_reserved()
        self.validate_config(config)
        self._policy_class = cls = self.get_policy_class()

        self.before_init()

        # Creating all workers (excluding evaluation workers).
        num_workers = config["num_workers"]
        self.workers = self._make_workers(
            env_creator=env_creator,
            validate_env=None,
            policy_class=cls,
            config=config,
            num_workers=num_workers,
        )
        self.train_exec_impl = self.execution_plan(self.workers, config)

        self.after_init()

        self.optimize_policy_backend()

    def restore_reserved(self) -> TrainerConfigDict:
        """Returns the final configuration."""
        restored = self._true_config
        del self._true_config
        return restored

    def validate_config(self, config: dict):
        """Assert final configurations are valid."""

    def get_policy_class(self) -> Type[Policy]:
        """Returns the Policy type to be set as the `_policy` attribute.

        Returns the `_policy` attribute by default. May be overriden to select a
        policy class depending on, e.g., the config dict.

        Called after :meth:`validate_config` and before :meth:`before_init`.
        """
        return self._policy_class

    def before_init(self):
        """Arbitrary setup before default initialization.

        Called after :meth:`get_policy_class` and before the creation of the
        worker set and execution plan.
        """

    @property
    def execution_plan(
        self,
    ) -> Callable[[WorkerSet, TrainerConfigDict], Iterable[ResultDict]]:
        """The execution plan function."""
        return default_execution_plan

    def after_init(self):
        """Arbitrary setup after default initialization.

        Called second-to-last in the :meth:`_init` procedure, after worker set
        and execution plan creation and before :meth:`optimize_policy_backend`.
        """

    def optimize_policy_backend(self):
        """Call `compile` on each policy if requested

        Called right after worker set creation. Requires the `compile` key under
        the `policy` subconfig.
        """
        if self.config["policy"].get("compile", False):
            self.workers.foreach_policy(lambda p, _: p.compile())

    @overrides(RLlibTrainer)
    def train(self) -> ResultDict:
        """Overrides super.train to synchronize global vars."""

        result = None
        for _ in range(1 + MAX_WORKER_FAILURE_RETRIES):
            try:
                result = Trainable.train(self)
            except RayError as err:
                if self.config["ignore_worker_failures"]:
                    logger.exception("Error in train call, attempting to recover")
                    self._try_recover()
                else:
                    logger.info(
                        "Worker crashed during call to train(). To attempt to "
                        "continue training without the failed worker, set "
                        "`'ignore_worker_failures': True`."
                    )
                    raise err
            except Exception as exc:
                time.sleep(0.5)  # allow logs messages to propagate
                raise exc
            else:
                break
        if result is None:
            raise RuntimeError("Failed to recover from worker crash")

        if hasattr(self, "workers") and isinstance(self.workers, WorkerSet):
            self._sync_filters_if_needed(self.workers)

        return result

    @overrides(RLlibTrainer)
    def step(self) -> dict:
        res = self.evaluate_pre_learning()
        res.update(next(self.train_exec_impl))
        res.update(self.evaluate_if_needed())
        return res

    def evaluate_pre_learning(self) -> dict:
        """Runs evaluation before any optimizations if requested.

        Returns:
            A dictionary with evaluation info
        """
        interval = self.config["evaluation_interval"]
        if self.iteration == 0 and interval and interval != 1:
            return self._evaluate()
        return {}

    def evaluate_if_needed(self) -> dict:
        """Runs evaluation episodes if configured to do so.

        Returns:
            A dictionary with evaluation info
        """
        iteration, interval = self.iteration + 1, self.config["evaluation_interval"]

        if interval == 1 or (iteration > 0 and interval and iteration % interval == 0):
            evaluation_metrics = self._evaluate()
            assert isinstance(
                evaluation_metrics, dict
            ), "_evaluate() needs to return a dict."
            return evaluation_metrics

        return {}

    @overrides(RLlibTrainer)
    def __getstate__(self) -> dict:
        state = super().__getstate__()
        state["train_exec_impl"] = self.train_exec_impl.shared_metrics.get().save()
        return state

    @overrides(RLlibTrainer)
    def __setstate__(self, state: dict):
        super().__setstate__(state)
        self.train_exec_impl.shared_metrics.get().restore(state["train_exec_impl"])

    @classmethod
    def default_resource_request(cls, config: PartialTrainerConfigDict) -> Resources:
        cnf = dict(cls.options.rllib_defaults, **config)
        cls._validate_config(cnf)
        num_workers = cnf["num_workers"] + cnf["evaluation_num_workers"]
        return Resources(
            cpu=cnf["num_cpus_for_driver"],
            gpu=cnf["num_gpus"],
            memory=cnf["memory"],
            object_store_memory=cnf["object_store_memory"],
            extra_cpu=cnf["num_cpus_per_worker"] * num_workers,
            extra_gpu=cnf["num_gpus_per_worker"] * num_workers,
            extra_memory=cnf["memory_per_worker"] * num_workers,
            extra_object_store_memory=cnf["object_store_memory_per_worker"]
            * num_workers,
        )



"""Losses for computing policy gradients."""
from typing import Optional, Tuple, Union

import torch
from nnrl.nn.actor import Alpha, DeterministicPolicy, StochasticPolicy
from nnrl.nn.critic import ClippedQValue, QValue, QValueEnsemble
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.utils.types import StatDict

from .abstract import Loss
from .utils import action_dpg, dist_params_stats


def clip_if_needed(critic: Union[QValue, QValueEnsemble]) -> QValue:
    if isinstance(critic, QValueEnsemble):
        critic = ClippedQValue(critic)
    return critic


class DeterministicPolicyGradient(Loss):
    """Loss function for Deterministic Policy Gradient.

    Args:
        actor: deterministic policy
        critic: action-value function (single or ensemble)
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)

    def __init__(
        self, actor: DeterministicPolicy, critic: Union[QValue, QValueEnsemble]
    ):
        self.actor = actor
        self.critic = clip_if_needed(critic)

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]
        act = self.actor(obs)
        val = self.critic(obs, act)
        loss = -torch.mean(val)

        stats = {"loss(actor)": loss.item()}
        return loss, stats


class ReparameterizedSoftPG(Loss):
    """Loss function for Soft Policy Iteration with reparameterized actor.

    Args:
        actor: stochastic reparameterized policy
        critic: action-value function (single or ensemble)
        alpha: entropy coefficient
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)

    def __init__(
        self,
        actor: StochasticPolicy,
        critic: Union[QValue, QValueEnsemble],
        alpha: Alpha,
    ):
        self.actor = actor
        self.critic = clip_if_needed(critic)
        self.alpha = alpha

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]

        action_values, entropy, stats = self.action_value_plus_entropy(obs)
        loss = -torch.mean(action_values + self.alpha() * entropy)

        stats.update({"loss(actor)": loss.item(), "entropy": entropy.mean().item()})
        return loss, stats

    def action_value_plus_entropy(self, obs: Tensor) -> Tuple[Tensor, Tensor, StatDict]:
        """
        Compute the action-value and a single sample estimate of the policy's entropy.
        """
        params = self.actor(obs)
        info = dist_params_stats(params, name="policy")

        act, logp = self.actor.dist.rsample(params)
        action_values = self.critic(obs, act)
        return action_values, -logp, info


class ActionDPG(Loss):
    """Deterministic Policy Gradient by separating action and Q-value grads.

    Args:
        actor: deterministic policy
        critic: Q-value function (single or ensemble)

    Attributes:
        dqda_clipping: Optional value by which to clip the action gradients
        clip_norm: Whether to clip action grads by norm or value
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)
    dqda_clipping: Optional[float] = None
    clip_norm: bool = True

    def __init__(
        self, actor: DeterministicPolicy, critic: Union[QValue, QValueEnsemble]
    ):
        self.actor = actor
        self.critic = clip_if_needed(critic)

    def compile(self):
        self.actor = torch.jit.script(self.actor)
        self.critic = torch.jit.script(self.critic)

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]
        a_max = self.actor(obs)
        q_max = self.critic(obs, a_max)

        loss, dqda_norm = action_dpg(q_max, a_max, self.dqda_clipping, self.clip_norm)
        loss = loss.mean()
        return loss, {"loss(actor)": loss.item(), "dqda_norm": dqda_norm.mean().item()}



# pylint:disable=missing-module-docstring
import numpy as np


def explained_variance(targets, pred):
    """Compute the explained variance given targets and predictions."""
    targets_var = np.var(targets, axis=0)
    diff_var = np.var(targets - pred, axis=0)
    return np.maximum(-1.0, 1.0 - (diff_var / targets_var))



# pylint:disable=missing-module-docstring
from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from ray.rllib import SampleBatch

from raylab.options import option
from raylab.utils.timer import TimerStat
from raylab.utils.types import StatDict

from ..stats import learner_stats


def model_based_options(cls: type) -> type:
    """Decorator to add default model-based options used by MBPolicyMixin."""
    model_update_interval = option(
        "model_update_interval",
        default=1,
        help="""Number of calls to `learn_on_batch` between each model update run.

        Example:
            With a 'rollout_fragment_length' of 1 and 'model_update_interval' of 25,
            will collect 25 environment transitions between each model optimization
            loop.
        """,
    )
    for opt in [model_update_interval]:
        cls = opt(cls)

    return cls


class MBPolicyMixin(ABC):
    """Off-policy mixin with dynamics model learning."""

    timers: Dict[str, TimerStat]
    _learn_calls: int = 0
    _info: dict

    def build_timers(self):
        """Create timers for model and policy training."""
        self.timers = {"model": TimerStat(), "policy": TimerStat()}
        self._info = {}

    @learner_stats
    def learn_on_batch(self, samples: SampleBatch) -> dict:
        # pylint:disable=missing-function-docstring
        self.add_to_buffer(samples)
        self._learn_calls += 1

        warmup = self._learn_calls == 1
        if (self._learn_calls % self.config["model_update_interval"] == 0) or warmup:
            with self.timers["model"] as timer:
                _, model_info = self.train_dynamics_model(warmup=warmup)
                timer.push_units_processed(model_info["model_epochs"])
                self._info.update(model_info)

        with self.timers["policy"] as timer:
            times = self.config["improvement_steps"]
            policy_info = self.update_policy(times=times)
            timer.push_units_processed(times)
            self._info.update(policy_info)

        self._info.update(self.timer_stats())
        return self._info.copy()

    @abstractmethod
    def train_dynamics_model(
        self, warmup: bool = False
    ) -> Tuple[List[float], StatDict]:
        """Implements the model training step.

        Calls the policy to optimize the model on the environment replay buffer.

        Args:
            warmup: Whether the optimization is being done on data collected
                via :meth:`sample_until_learning_starts`.

        Returns:
            A tuple containing the list of evaluation losses for each model and
            a dictionary of training statistics
        """

    def update_policy(self, times: int) -> StatDict:
        """Improve the policy on previously collected environment data.

        Improves the policy on batches sampled from the replay buffer.

        Args:
            times: number of times to call :meth:`improve_policy`

        Returns:
            A dictionary of training statistics
        """
        for _ in range(times):
            batch = self.replay.sample(self.config["batch_size"])
            batch = self.lazy_tensor_dict(batch)
            info = self.improve_policy(batch)

        return info

    def timer_stats(self) -> dict:
        """Returns the timer statistics."""
        model_timer = self.timers["model"]
        policy_timer = self.timers["policy"]
        return dict(
            model_time_s=round(model_timer.mean, 3),
            policy_time_s=round(policy_timer.mean, 3),
            # Get mean number of model epochs per second spent updating the model
            model_update_throughput=round(model_timer.mean_throughput, 3),
            # Get mean number of policy updates per second spent updating the policy
            policy_update_throughput=round(policy_timer.mean_throughput, 3),
        )

    @staticmethod
    def add_options(policy_cls: type) -> type:
        """Decorator to add default model-based options used by MBPolicyMixin."""
        return model_based_options(policy_cls)



# pylint:disable=missing-module-docstring
from __future__ import annotations

import copy
import statistics as stats
from dataclasses import dataclass, field
from typing import Any, Optional, OrderedDict

import pytorch_lightning as pl
import torch
from dataclasses_json import DataClassJsonMixin
from nnrl.nn.model import SME
from nnrl.types import TensorDict
from nnrl.utils import convert_to_tensor
from torch import Tensor, nn
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset, random_split

from raylab.options import option
from raylab.policy.losses import Loss
from raylab.utils.lightning import lightning_warnings_only, suppress_dataloader_warnings
from raylab.utils.replay_buffer import NumpyReplayBuffer
from raylab.utils.types import StatDict

# ======================================================================================
# LightningModel
# ======================================================================================


class LightningModel(pl.LightningModule):
    # pylint:disable=too-many-ancestors,arguments-differ,missing-docstring
    train_loss: Loss
    val_loss: Loss
    test_loss: Loss
    early_stop_on = "early_stop_on"

    def __init__(self, model: nn.Module, loss: Loss, optimizer: Optimizer):
        super().__init__()
        self.model = model
        self.configure_losses(loss)
        self.optimizer = optimizer

    def configure_optimizers(self) -> Optimizer:
        return self.optimizer

    def configure_losses(self, loss: Loss):
        self.train_loss = self.val_loss = self.test_loss = loss

    def forward(self, batch: TensorDict) -> tuple[Tensor, StatDict]:
        return self.train_loss(batch)

    def training_step(self, batch: TensorDict, _) -> Tensor:
        loss, info = self.train_loss(batch)
        info = self.stat_to_tensor_dict(info)
        self.log("train/loss", loss)
        self.log(self.early_stop_on, loss)
        self.log_dict({"train/" + k: v for k, v in info.items()})
        return loss

    def validation_step(self, batch: TensorDict, _) -> Tensor:
        loss, info = self.val_loss(batch)
        info = self.stat_to_tensor_dict(info)
        self.log("val/loss", loss)
        self.log(self.early_stop_on, loss)
        self.log_dict({"val/" + k: v for k, v in info.items()})
        return loss

    def test_step(self, batch: TensorDict, _) -> Tensor:
        loss, info = self.test_loss(batch)
        info = self.stat_to_tensor_dict(info)
        self.log("test/loss", loss)
        self.log_dict({"test/" + k: v for k, v in info.items()})
        return loss

    def stat_to_tensor_dict(self, info: StatDict) -> TensorDict:
        return {k: convert_to_tensor(v, self.device) for k, v in info.items()}


# ======================================================================================
# DataModule
# ======================================================================================


@dataclass
class DatamoduleSpec(DataClassJsonMixin):
    """Specifications for creating the data module.

    Attributes:
        holdout_ratio: Fraction of replay buffer to use as validation dataset
        max_holdout: Maximum number of samples to use as validation dataset
        batch_size: Size of minibatch for dynamics model training
        shuffle: set to ``True`` to have the data reshuffled
            at every epoch (default: ``True``).
        num_workers: How many subprocesses to use for data loading.
            ``0`` means that the data will be loaded in the main process.
    """

    holdout_ratio: float = 0.2
    max_holdout: Optional[int] = None
    batch_size: int = 64
    shuffle: bool = True
    num_workers: int = 0  # Use at least one worker for speedup

    def __post_init__(self):
        assert self.holdout_ratio < 1.0, "Holdout data cannot be the entire dataset"
        assert (
            not self.max_holdout or self.max_holdout >= 0
        ), "Maximum number of holdout samples must be non-negative"
        assert self.batch_size > 0, "Model batch size must be positive"


class DataModule(pl.LightningDataModule):
    """Data module from experience replay buffer

    Args:
        replay: Experience replay buffer
        spec: Data loading especifications
    """

    # pylint:disable=abstract-method
    train_dataset: Dataset
    val_dataset: Dataset

    def __init__(self, replay: NumpyReplayBuffer, spec: DatamoduleSpec):
        assert isinstance(replay, NumpyReplayBuffer)
        super().__init__()
        self.replay_dataset = ReplayDataset(replay)
        self.spec = spec

    def setup(self, stage: Optional[str] = None):
        dataset = self.replay_dataset
        spec = self.spec
        replay_count = len(dataset)
        max_holdout = spec.max_holdout or replay_count
        val_size = min(round(replay_count * spec.holdout_ratio), max_holdout)
        self.train_dataset, self.val_dataset = random_split(
            dataset, (replay_count - val_size, val_size)
        )

    def train_dataloader(self):
        spec = self.spec
        kwargs = dict(
            shuffle=spec.shuffle,
            batch_size=spec.batch_size,
            num_workers=spec.num_workers,
        )
        return DataLoader(self.train_dataset, **kwargs)

    def val_dataloader(self):
        if len(self.val_dataset) == 0:
            return None

        spec = self.spec
        kwargs = dict(
            shuffle=False, batch_size=spec.batch_size, num_workers=spec.num_workers
        )
        return DataLoader(self.val_dataset, **kwargs)


class ReplayDataset(Dataset):
    """Adapter for using a replay buffer as an map-style dataset."""

    def __init__(self, replay: NumpyReplayBuffer):
        self.replay = replay

    def __len__(self):
        return len(self.replay)

    def __getitem__(self, idx: int):
        return self.replay[idx]


# ======================================================================================
# EarlyStopping
# ======================================================================================


class EarlyStopping(pl.callbacks.EarlyStopping):
    # pylint:disable=missing-docstring
    _train_outputs: list[tuple[Tensor, StatDict]]
    _val_outputs: list[tuple[Tensor, StatDict]]
    _loss: tuple[list[float], StatDict] = None
    _module_state: Optional[OrderedDict[str, Tensor]] = None

    def setup(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        stage: Optional[str] = None,
    ):
        super().setup(trainer, pl_module, stage)
        self._train_outputs = []
        self._val_outputs = []

    def on_train_epoch_start(self, trainer, pl_module):
        self._train_outputs = []
        super().on_train_epoch_start(trainer, pl_module)

    def on_train_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        # pylint:disable=too-many-arguments
        self._train_outputs += [pl_module.val_loss.last_output]
        super().on_train_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )

    def on_validation_epoch_start(self, trainer, pl_module):
        self._val_outputs = []
        super().on_validation_epoch_start(trainer, pl_module)

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        # pylint:disable=too-many-arguments
        self._val_outputs += [pl_module.val_loss.last_output]
        super().on_validation_batch_end(
            trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        )

    def _run_early_stopping_check(self, trainer):
        if self.patience is None:
            # Always save latest outputs
            self.save_outputs()
        else:
            super()._run_early_stopping_check(trainer)
            # Save outputs only if improved or none have been logged yet
            if self.wait_count == 0:  # Improved
                self.save_outputs()
                self.save_module_state(trainer.lightning_module)
            elif self._loss is None:
                self.save_outputs()

    def save_outputs(self):
        # Give preference to validation outputs
        epoch_outputs = self._val_outputs or self._train_outputs

        epoch_losses, epoch_infos = zip(*epoch_outputs)
        model_losses = torch.stack(epoch_losses, dim=0).mean(dim=0).tolist()
        model_infos = {k: stats.mean(i[k] for i in epoch_infos) for k in epoch_infos[0]}
        self._loss = (model_losses, model_infos)

    def save_module_state(self, pl_module: nn.Module):
        self._module_state = copy.deepcopy(pl_module.state_dict())

    @property
    def loss(self) -> tuple[list[float], StatDict]:
        return self._loss

    @property
    def module_state(self) -> OrderedDict[str, Tensor]:
        return self._module_state

    def on_save_checkpoint(
        self, trainer, pl_module, checkpoint: dict[str, Any]
    ) -> dict[str, Any]:
        state = super().on_save_checkpoint(trainer, pl_module, checkpoint)
        state.update(loss=self._loss, module=self._module_state)
        return state

    def on_load_checkpoint(self, callback_state: dict[str, Any]):
        state_dict = callback_state
        self._loss = state_dict["loss"]
        self._module_state = copy.deepcopy(state_dict["module"])
        used = set("loss module".split())
        super().on_load_checkpoint(
            {k: v for k, v in state_dict.items() if k not in used}
        )


# ======================================================================================
# Model Trainer
# ======================================================================================


@dataclass
class LightningTrainerSpec(DataClassJsonMixin):
    """Specifications for Lightning trainers.

    Attributes:
        max_epochs: Maximum number of full model passes through the data
        max_steps: Maximum number of model gradient steps
        patience: Tolerate this many epochs of successive performance
            degradation. If None, disables early stopping.
        improvement_delta: Minimum expected absolute improvement in model
            validation loss
    """

    max_epochs: Optional[int] = 1
    max_steps: Optional[int] = None
    patience: Optional[int] = 1
    improvement_delta: float = 0.0

    def __post_init__(self):
        if self.max_epochs is None:
            self.max_epochs = self.max_steps

        assert self.max_epochs > 0, "Maximum number of epochs must be positive"
        assert not self.max_steps or self.max_steps > 0
        assert (
            self.patience is None or self.patience >= 0
        ), "Patience must be nonnegative or None"
        assert isinstance(
            self.improvement_delta, float
        ), "Improvement threshold must be a scalar"

    def build_trainer(
        self, check_val: bool, check_on_train_epoch_end: bool
    ) -> tuple[pl.Trainer, EarlyStopping]:
        """Returns the Pytorch Lightning configured with this spec."""
        early_stopping = EarlyStopping(
            monitor=LightningModel.early_stop_on,
            min_delta=self.improvement_delta,
            patience=self.patience,
            mode="min",
            strict=True,
            check_on_train_epoch_end=check_on_train_epoch_end,
        )
        trainer = pl.Trainer(
            logger=False,
            num_sanity_val_steps=2 if check_val else 0,
            callbacks=[early_stopping],
            max_epochs=self.max_epochs,
            max_steps=self.max_steps,
            track_grad_norm=2,
            progress_bar_refresh_rate=0,  # don't show progress bar for model training
            weights_summary=None,  # don't print summary before training
            checkpoint_callback=False,  # don't save last model
        )
        return trainer, early_stopping


@dataclass
class TrainingSpec(DataClassJsonMixin):
    """Specifications for training the model.

    Attributes:
        datamodule: Specifications for creating the data module
        training: Specifications for model training
        warmup: Specifications for model warmup
    """

    datamodule: DatamoduleSpec = field(default_factory=DatamoduleSpec)
    training: LightningTrainerSpec = field(default_factory=LightningTrainerSpec)
    warmup: LightningTrainerSpec = field(default_factory=LightningTrainerSpec)


class LightningModelTrainer:
    """Model training behavior for TorchPolicy instances via PyTorch Lightning.

    Args:
        models: Stochastic model ensemble
        loss_fn: Loss associated with the model ensemble
        optimizer: Optimizer associated with the model ensemble
        replay: Experience replay buffer
        config: Dictionary containing `model_training` and `model_warmup` dicts

    Attributes:
        pl_model: Pytorch Lightning model
        datamodule: Lightning data module
        spec: Specifications for training the model
        training_loss: Loss function used for model training and evaluation
        warmup_loss: Loss function used for model warm-up.
    """

    pl_model: LightningModel
    datamodule: DataModule
    spec: TrainingSpec

    def __init__(
        self,
        models: SME,
        loss_fn: Loss,
        optimizer: Optimizer,
        replay: NumpyReplayBuffer,
        config: dict,
    ):
        # pylint:disable=too-many-arguments
        self.spec = TrainingSpec.from_dict(config["model_training"])
        self.pl_model = LightningModel(model=models, loss=loss_fn, optimizer=optimizer)
        self.datamodule = DataModule(replay, self.spec.datamodule)
        self.training_loss = self.warmup_loss = loss_fn

    def optimize(self, warmup: bool = False) -> tuple[list[float], StatDict]:
        """Update models using replay buffer data.

        Args:
            warmup: Whether to train with warm-up loss and spec

        Returns:
            A tuple with a list of each model's evaluation loss and a dictionary
            with training statistics
        """
        loss_fn = self.warmup_loss if warmup else self.training_loss
        self.pl_model.configure_losses(loss_fn)

        trainer_spec = self.spec.warmup if warmup else self.spec.training
        trainer, early_stopping = trainer_spec.build_trainer(
            check_val=warmup,
            check_on_train_epoch_end=self.spec.datamodule.holdout_ratio == 0.0,
        )

        self.run_training(
            model=self.pl_model, trainer=trainer, datamodule=self.datamodule
        )
        losses, info = early_stopping.loss
        info.update(self.trainer_info(trainer))
        self.check_early_stopping(early_stopping, self.pl_model)
        return losses, info

    @staticmethod
    @lightning_warnings_only()
    @suppress_dataloader_warnings()
    def run_training(
        model: LightningModel, trainer: pl.Trainer, datamodule: DataModule
    ):
        """Trains model and handles checkpointing."""
        trainer.fit(model, datamodule=datamodule)

    @staticmethod
    def trainer_info(trainer: pl.Trainer) -> dict:
        """Returns a dictionary of training run info."""
        return dict(
            model_epochs=trainer.current_epoch + 1, model_steps=trainer.global_step
        )

    @staticmethod
    def check_early_stopping(early_stopping: EarlyStopping, model: nn.Module):
        """Restore best model parameters if training was early stopped."""
        saved_state = early_stopping.module_state
        if saved_state:
            model.load_state_dict(saved_state)

    @staticmethod
    def add_options(cls_: type) -> type:
        """Add options for classes that may use this class."""
        return option(
            "model_training",
            default=TrainingSpec().to_dict(),
            help=TrainingSpec.__doc__,
        )(cls_)



# pylint:disable=missing-module-docstring
from ray.util.timer import _Timer


class TimerStat(_Timer):
    """Ray timer that returns itself upon entering a context."""

    def __enter__(self) -> "TimerStat":
        super().__enter__()
        return self



"""Common utilities for Loss functions."""
from typing import Optional, Tuple

import torch
from nnrl.types import TensorDict
from torch import Tensor
from torch.autograd import grad

from raylab.utils.types import StatDict


def action_dpg(
    q_max: Tensor,
    a_max: Tensor,
    dqda_clipping: Optional[float] = None,
    clip_norm: bool = False,
) -> Tuple[Tensor, Tensor]:
    # pylint:disable=line-too-long
    """Deterministic policy gradient loss, similar to trfl.dpg.

    Inspired by `Acmes's DPG`_. Allows logging of Q-value action gradients.

    .. _`Acme's DPG`: https://github.com/deepmind/acme/blob/51c4db7c8ec27e040ac52d65347f6f4ecfe04f81/acme/tf/losses/dpg.py#L21

    Args:
        q_max: Q-value of the approximate greedy action of shape `(*,)`
        a_max: Action from the policy's output of shape `(*, A)`
        dqda_clipping: Optional value by which to clip the action gradients
        clip_norm: Whether to clip action grads by norm or value

    Returns:
        The DPG loss of shape `(*,)` and the norm of the action-value gradient
        of shape `(*,)`

    Note:
        This is a loss, so it's already supposed to be minimized, i.e., its
        gradients w.r.t. the policy parameters correspond to the negative DPG
    """
    # pylint:enable=line-too-long
    # Fake a Jacobian-vector product to calculate grads w.r.t. to batch of actions
    (dqda,) = grad(q_max, a_max, grad_outputs=torch.ones_like(q_max))
    # Keepdim to support broadcasting in clip_norm later
    dqda_norm = torch.norm(dqda, dim=-1, keepdim=True)

    if dqda_clipping:
        # pylint:disable=invalid-unary-operand-type
        if clip_norm:
            clip_coef = dqda_clipping / dqda_norm
            dqda = torch.where(clip_coef < 1, dqda * clip_coef, dqda)
        else:
            dqda = torch.clamp(dqda, min=-dqda_clipping, max=dqda_clipping)

    loss = -torch.sum(a_max * dqda.detach(), dim=-1)
    return loss, dqda_norm


def dist_params_stats(dist_params: TensorDict, name: str) -> StatDict:
    """Returns mean, max, and min for each distribution parameter.

    Args:
        dist_params: Dictionary mapping names to distribution parameters

    Returns:
        Dictionary with average, minimum, and maximum of each parameter as
        floats
    """
    items = tuple((k, v) for k, v in dist_params.items() if v.requires_grad)
    info = {}
    info.update({name + "/mean_" + k: v.mean().item() for k, v in items})
    info.update({name + "/max_" + k: v.max().item() for k, v in items})
    info.update({name + "/min_" + k: v.min().item() for k, v in items})
    return info



"""Utilities for manipulating python dictionaries and general mappings."""
import copy
from typing import Optional

from ray.rllib.utils import deep_update


def deep_merge(
    dict1,
    dict2,
    new_keys_allowed: bool = False,
    allow_new_subkey_list: Optional[list] = None,
    override_all_if_type_changes: Optional[list] = None,
):
    """Deep copy original dict and pass it to RLlib's deep_update."""
    clone = copy.deepcopy(dict1)
    return deep_update(
        clone,
        dict2,
        new_keys_allowed=new_keys_allowed,
        allow_new_subkey_list=allow_new_subkey_list,
        override_all_if_type_changes=override_all_if_type_changes,
    )


def get_keys(mapping, *keys):
    """Return the values corresponding to the given keys, in order."""
    return (mapping[k] for k in keys)


def all_except(mapping, *exclude):
    """Return a new mapping with all keys except `exclude`.

    Keys must be hashable to be used with `set`.
    """
    exclude = set(exclude)
    return {k: v for k, v in mapping.items() if k not in exclude}



"""Experiment monitoring with Streamlit."""
from __future__ import annotations

from typing import Any

import streamlit as st

from raylab.cli.viz import time_series
from raylab.utils import exp_data as exp_util

# pylint:disable=invalid-name,missing-docstring,pointless-string-statement
# pylint:disable=no-value-for-parameter,unsubscriptable-object
"""
# Raylab
"""


# https://discuss.streamlit.io/t/how-can-i-clear-a-specific-cache-only/1963/6
@st.cache(allow_output_mutation=True)
def load_data(
    directories: list[str], include_errors: bool = False
) -> list[list[exp_util.ExperimentData]]:
    return [exp_util.load_exps_data(directories, include_errors=include_errors)]


@st.cache
def get_exp_root_folders(directories: list[str]) -> list[exp_util.Folder]:
    return exp_util.get_folders_with_target_files(directories, is_experiment_root)


def is_experiment_root(path: str) -> bool:
    return path.startswith("experiment_state") and path.endswith(".json")


def dict_value_multiselect(mapping, name=None):
    items = []

    keys = st.multiselect(f"{name}:", list(mapping.keys()), key=name)
    if keys:
        for key in keys:
            choices = list(mapping[key])
            values = st.multiselect(f"{key} values:", choices, key=name)
            for val in values:
                items.append((key, val))

    return items


def add_sidebar_options() -> dict:
    return {
        "include_errors": st.sidebar.checkbox("Include experiments with errors"),
        "legend_location": st.sidebar.selectbox(
            "legend location",
            """
            top_left
            top_center
            top_right
            center_right
            bottom_right
            bottom_center
            bottom_left
            center_left
            center
            """.split(),
        ),
        "legend_orientation": st.sidebar.selectbox(
            "legend orientation", "vertical horizontal".split()
        ),
        "individual": st.sidebar.checkbox("Show individual curves"),
        "standard_error": st.sidebar.checkbox("Use standard error"),
        "log_scale": st.sidebar.checkbox("Log scale"),
    }


def get_selector(folders: list[str], sidebar_options: dict) -> exp_util.Selector:
    data_wrapper = load_data(
        list(folders), include_errors=sidebar_options["include_errors"]
    )
    if st.button("Reload Data"):
        data_wrapper.clear()
        data_wrapper.append(
            exp_util.load_exps_data(
                tuple(folders), include_errors=sidebar_options["include_errors"]
            )
        )
    exps_data = data_wrapper[0]
    distinct_params = dict(sorted(exp_util.extract_distinct_params(exps_data)))

    include = dict_value_multiselect(distinct_params, name="Include")
    exclude = dict_value_multiselect(distinct_params, name="Exclude")

    [selector], _ = exp_util.filter_and_split_experiments(
        exps_data, include=include, exclude=exclude
    )
    return selector


def select_axis_keys(exps_data):
    plottable_keys = exp_util.get_plottable_keys(exps_data)
    x_plottable_keys = exp_util.get_x_plottable_keys(plottable_keys, exps_data)
    x_key = st.selectbox(
        "X axis:",
        x_plottable_keys,
        index=safe_index(x_plottable_keys, "timesteps_total"),
    )
    y_key = st.selectbox(
        "Y axis:",
        plottable_keys,
        index=safe_index(plottable_keys, "episode_reward_mean"),
    )
    return x_key, y_key


def safe_index(seq: list, value: Any) -> int:
    index = 0
    try:
        index = seq.index(value)
    except ValueError:
        pass
    return index


def split_and_label_groups(exps_data, selector):
    distinct_params = dict(sorted(exp_util.extract_distinct_params(exps_data)))
    split = st.selectbox(
        "Group by:",
        [""] + list(distinct_params.keys()),
        format_func=lambda x: "(none)" if x == "" else x,
    )
    if split:
        values = distinct_params.get(split, [])
        labels = list(map(str, values))
        groups = [selector.where(split, v) for v in values]
    else:
        labels = ["experiment"]
        groups = [selector]

    return groups, labels, split


def main():
    import sys

    directories = tuple(sys.argv[1:])
    root_exp_folders = [f.path for f in get_exp_root_folders(directories)]
    folders = st.sidebar.multiselect(
        "Filter experiments:", root_exp_folders, default=root_exp_folders
    )
    sidebar_options = add_sidebar_options()

    if folders:
        selector = get_selector(folders, sidebar_options)
        exps_data = selector.extract()
        if exps_data:
            x_key, y_key = select_axis_keys(exps_data)
            groups, labels, split = split_and_label_groups(exps_data, selector)

            chart = time_series(x_key, y_key, groups, labels, sidebar_options)

            chart.legend.title = split
            chart.legend.title_text_font_style = "bold"
            chart.legend.location = sidebar_options["legend_location"]
            chart.legend.orientation = sidebar_options["legend_orientation"]
            st.bokeh_chart(chart)


if __name__ == "__main__":
    main()



"""RAYLAB: Extensions and custom algorithms in RLlib."""
import poetry_version

__author__ = """Ângelo Gregório Lovatto"""
__email__ = "angelolovatto@gmail.com"
__version__ = poetry_version.extract(source_file=__file__)


def register_all_agents():
    """Register all trainer names in Tune."""
    from ray.tune import register_trainable

    from raylab.agents.registry import AGENTS

    for name, trainer_import in AGENTS.items():
        register_trainable(name, trainer_import())


def register_all_environments():
    """Register all custom environments in Tune."""
    from ray.tune import register_env

    from raylab.envs.registry import ENVS

    for name, env_creator in ENVS.items():
        register_env(name, env_creator)


def register_all():
    """Register all trainers and environments in Tune."""
    register_all_agents()
    register_all_environments()



# pylint:disable=missing-module-docstring
import logging
from typing import List

from ray.rllib import SampleBatch
from ray.rllib.execution.common import (
    STEPS_SAMPLED_COUNTER,
    _check_sample_batch_type,
    _get_shared_metrics,
)
from ray.rllib.utils.typing import SampleBatchType

logger = logging.getLogger(__name__)


class LearningStarts:
    """Callable used to merge initial batches until a desired timestep count is met

    This should be used with the .combine() operator.

    Examples:
        >>> rollouts = ParallelRollouts(...)
        >>> rollouts = rollouts.combine(LearningStarts(learning_starts=10000))
        >>> print(next(rollouts).count)
        10000
    """

    # pylint:disable=too-few-public-methods
    def __init__(self, learning_starts: int):
        self.learning_starts = learning_starts
        self.buffer = []
        self.count = 0
        self.done = False

    def __call__(self, batch: SampleBatchType) -> List[SampleBatchType]:
        _check_sample_batch_type(batch)
        if self.done:
            # Warmup phase done, simply return batch
            return [batch]

        metrics = _get_shared_metrics()
        timesteps_total = metrics.counters[STEPS_SAMPLED_COUNTER]
        self.buffer.append(batch)
        self.count += batch.count
        assert self.count == timesteps_total

        if timesteps_total < self.learning_starts:
            # Return emtpy if still in warmup
            return []

        # Warmup just done
        if self.count > self.learning_starts * 2:
            logger.info(  # pylint:disable=logging-fstring-interpolation
                "Collected more training samples than expected "
                f"(actual={self.count}, expected={self.learning_starts}). "
                "This may be because you have many workers or "
                "long episodes in 'complete_episodes' batch mode."
            )
        out = SampleBatch.concat_samples(self.buffer)
        self.buffer = []
        self.count = 0
        self.done = True
        return [out]



"""Utilities for visualization."""
from __future__ import annotations

import bokeh
import numpy as np
import pandas as pd
from bokeh.models import (
    BoxZoomTool,
    ColumnDataSource,
    HelpTool,
    HoverTool,
    PanTool,
    ResetTool,
    SaveTool,
    WheelZoomTool,
)
from bokeh.plotting import figure

from raylab.utils.exp_data import ExperimentData, Selector


def time_series(
    x_key: str, y_key: str, groups: list[Selector], labels: list[str], config: dict
):
    """Plot time series with error bands per group."""
    # pylint:disable=too-many-function-args
    kwargs = dict(y_axis_type="log") if config["log_scale"] else {}
    pic = figure(title="Plot", **kwargs)
    pic.xaxis.axis_label = x_key
    pic.yaxis.axis_label = y_key

    pic.tools = [
        PanTool(),
        BoxZoomTool(),
        WheelZoomTool(dimensions="height"),
        WheelZoomTool(dimensions="width"),
        SaveTool(),
        ResetTool(),
        HelpTool(),
    ]
    pic.add_tools()
    if config["individual"]:
        pic.add_tools(HoverTool(tooltips=[("y", "@y"), ("x", "@x{a}"), ("id", "@id")]))
    else:
        pic.add_tools(HoverTool(tooltips=[("y", "@y_mean"), ("x", "@x{a}")]))

    for label, group, color in zip(labels, groups, bokeh.palettes.cividis(len(labels))):
        data = group.extract()
        progresses = [d.progress for d in data if y_key in d.progress.columns]
        if not progresses:
            continue

        x_all, all_ys = filter_and_interpolate(x_key, y_key, progresses)

        if config["individual"]:
            plot_individual(pic, x_all, all_ys, data, label, color)
        else:
            plot_mean_dispersion(
                pic,
                x_all,
                all_ys,
                label,
                color,
                standard_error=config["standard_error"],
            )

    pic.legend.location = "bottom_left"
    pic.legend.click_policy = "hide"
    return pic


def filter_and_interpolate(
    x_key: str, y_key: str, progresses: list[pd.DataFrame]
) -> tuple[np.ndarray, list[np.ndarray]]:
    # pylint:disable=missing-function-docstring
    # Filter NaN values from plots
    masks = [~np.isnan(p[y_key]) for p in progresses]
    xs_ = [p[x_key][m] for m, p in zip(masks, progresses)]
    ys_ = [p[y_key][m] for m, p in zip(masks, progresses)]
    x_all = np.unique(np.sort(np.concatenate(xs_)))
    all_ys = [
        np.interp(x_all, x, y, left=np.nan, right=np.nan) for x, y in zip(xs_, ys_)
    ]
    return x_all, all_ys


def plot_individual(
    pic,
    x_all: np.ndarray,
    all_ys: list[np.ndarray],
    data: ExperimentData,
    label: str,
    color: str,
):
    # pylint:disable=missing-function-docstring,too-many-arguments
    for datum, y_i in zip(data, all_ys):
        identifier = str(datum.params["id"])
        dataframe = pd.DataFrame({"x": x_all, "y": y_i, "id": identifier})
        source = ColumnDataSource(data=dataframe)
        pic.line(
            x="x",
            y="y",
            source=source,
            legend_label=label,
            color=color,
        )


def plot_mean_dispersion(
    pic,
    x_all: np.ndarray,
    all_ys: list[np.ndarray],
    label: str,
    color: str,
    standard_error: bool = False,
):
    # pylint:disable=missing-function-docstring,too-many-arguments
    y_mean = np.nanmean(all_ys, axis=0)
    dispersion = np.nanstd(all_ys, axis=0)
    if standard_error:
        dispersion /= np.sqrt(np.sum(1 - np.isnan(y_mean), axis=0))
    dataframe = pd.DataFrame(
        {
            "x": x_all,
            "y_mean": y_mean,
            "y_low": y_mean - dispersion,
            "y_high": y_mean + dispersion,
            "label": label,
        }
    )
    source = ColumnDataSource(data=dataframe)
    pic.line(x="x", y="y_mean", source=source, legend_label=label, color=color)
    pic.varea(
        x="x",
        y1="y_low",
        y2="y_high",
        source=source,
        fill_alpha=0.25,
        legend_label=label,
        color=color,
    )



"""SVG(1) policy class using PyTorch."""
import contextlib
import warnings

import torch
from nnrl.optim import get_optimizer_class
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import Tensor, nn

from raylab.agents.svg import SVGTorchPolicy
from raylab.options import configure, option
from raylab.policy import AdaptiveKLCoeffMixin, EnvFnMixin, TorchPolicy, learner_stats
from raylab.policy.losses import OneStepSVG
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options
from raylab.utils.replay_buffer import ReplayField


@configure
@off_policy_options
@option("optimizer/type", "Adam", help="Optimizer type for model, actor, and critic")
@option("optimizer/model", {"lr": 1e-3})
@option("optimizer/actor", {"lr": 1e-3})
@option("optimizer/critic", {"lr": 1e-3})
@option("max_grad_norm", 10.0, help="Clip gradient norms by this value")
@option(
    "replay_kl",
    True,
    help="""
    Whether to penalize KL divergence with the current policy or past policies
    that generated the replay pool.
    """,
)
@option(
    "kl_schedule",
    {"initial_coeff": 0},
    help="Options for adaptive KL coefficient. See raylab.utils.adaptive_kl",
)
@option("module/type", default="SVG")
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
@option("exploration_config/pure_exploration_steps", 1000)
class SVGOneTorchPolicy(OffPolicyMixin, AdaptiveKLCoeffMixin, SVGTorchPolicy):
    """Stochastic Value Gradients policy for off-policy learning."""

    # pylint:disable=too-many-ancestors

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_actor = OneStepSVG(
            self.module.model, self.module.actor, self.module.critic
        )
        self.loss_actor.gamma = self.config["gamma"]

        self.build_replay_buffer()

    def build_replay_buffer(self):
        super().build_replay_buffer()
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))

    @override(TorchPolicy)
    def compile(self):
        warnings.warn(f"{type(self).__name__} is incompatible with TorchScript")

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_module(self, obs_space, action_space, config: dict):
        config["module"].setdefault("actor", {})
        config["module"]["actor"]["old_policy"] = config["replay_kl"]
        return super()._make_module(obs_space, action_space, config)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        """PyTorch optimizer to use."""
        optimizers = super()._make_optimizers()
        options = self.config["optimizer"]
        cls = get_optimizer_class(options["type"], wrap=True)
        modules = {
            "model": self.module.model,
            "actor": self.module.actor,
            "critic": self.module.critic,
        }
        param_groups = [
            dict(params=mod.parameters(), **options[name])
            for name, mod in modules.items()
        ]
        optimizers["all"] = cls(param_groups)
        return optimizers

    @learner_stats
    @override(OffPolicyMixin)
    def learn_on_batch(self, samples: SampleBatch) -> dict:
        self.update_old_policy()
        info = super().learn_on_batch(samples)
        info.update(self.update_kl_coeff(samples))
        return info

    def update_old_policy(self):
        """Copy params of current policy into old one for future KL computation."""
        self.module.old_actor.load_state_dict(self.module.actor.state_dict())

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict) -> dict:
        batch, info = self.add_truncated_importance_sampling_ratios(batch)

        with self.optimizers.optimize("all"):
            model_value_loss, stats = self.compute_joint_model_value_loss(batch)
            info.update(stats)
            model_value_loss.backward()

            with self.freeze_model_and_critic():
                svg_loss, stats = self.loss_actor(batch)
                info.update(stats)
                kl_loss = self.curr_kl_coeff * self._avg_kl_divergence(batch)
                (svg_loss + kl_loss).backward()

        info.update(self.extra_grad_info(batch))
        self._update_polyak()
        return info

    @contextlib.contextmanager
    def freeze_model_and_critic(self):
        """Disable gradients for model and critic."""
        self.module.model.requires_grad_(False)
        self.module.critic.requires_grad_(False)
        yield
        self.module.model.requires_grad_(True)
        self.module.critic.requires_grad_(True)

    @torch.no_grad()
    @override(AdaptiveKLCoeffMixin)
    def _kl_divergence(self, sample_batch):
        batch = self.lazy_tensor_dict(sample_batch)
        return self._avg_kl_divergence(batch).item()

    def _avg_kl_divergence(self, batch: TensorDict) -> Tensor:
        if self.config["replay_kl"]:
            logp = self.module.actor.log_prob(
                batch[SampleBatch.CUR_OBS], batch[SampleBatch.ACTIONS]
            )
            return torch.mean(batch[SampleBatch.ACTION_LOGP] - logp)

        old_act, old_logp = self.module.old_actor.rsample(batch[SampleBatch.CUR_OBS])
        logp = self.module.actor.log_prob(batch[SampleBatch.CUR_OBS], old_act)
        return torch.mean(old_logp - logp)

    @torch.no_grad()
    def extra_grad_info(self, batch: TensorDict) -> dict:
        """Compute gradient norms and policy statistics."""
        grad_norms = {
            f"grad_norm({k})": nn.utils.clip_grad_norm_(
                getattr(self.module, k).parameters(), float("inf")
            )
            for k in "model actor critic".split()
        }
        policy_info = {
            "entropy": self.module.actor.log_prob(
                batch[SampleBatch.CUR_OBS], batch[SampleBatch.ACTIONS]
            )
            .mean()
            .neg()
            .item(),
            "curr_kl_coeff": self.curr_kl_coeff,
        }
        return {**grad_norms, **policy_info}



# pylint:disable=missing-module-docstring
from typing import Tuple, Union

import torch
from nnrl.nn.actor import DeterministicPolicy, StochasticPolicy
from nnrl.nn.critic import QValueEnsemble, VValue
from nnrl.nn.model import SME, StochasticModel
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor, nn

from raylab.utils.types import StatDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin, UniformModelPriorMixin
from .q_learning import QLearningMixin
from .utils import dist_params_stats


class StochasticPolicyAction(nn.Module):
    # pylint:disable=missing-docstring,arguments-differ,abstract-method
    def __init__(self, policy: StochasticPolicy):
        super().__init__()
        self.policy = policy

    def forward(self, obs: Tensor) -> Tensor:
        sample, _ = self.policy.sample(obs)
        return sample


class DeterministicPolicyAction(nn.Module):
    # pylint:disable=missing-docstring,arguments-differ,abstract-method
    def __init__(self, policy: DeterministicPolicy):
        super().__init__()
        self.policy = policy

    def forward(self, obs: Tensor) -> Tensor:
        return self.policy(obs)


class DynaQLearning(UniformModelPriorMixin, EnvFunctionsMixin, Loss):
    """Loss function Dyna-augmented Clipped Double Q-learning.

    Attributes:
        critics: Main action-values
        actor: Stochastic or deterministic policy
        models: Stochastic model ensemble
        target_critic: Target state-value function
        batch_keys: Keys required to be in the tensor batch
        gamma: discount factor
    """

    critics: QValueEnsemble
    actor: Union[DeterministicPolicy, StochasticPolicy]
    models: Union[StochasticModel, SME]
    target_critic: VValue

    gamma: float = 0.99
    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)
    _model_samples: int = 1

    def __init__(
        self,
        critics: QValueEnsemble,
        actor: Union[DeterministicPolicy, StochasticPolicy],
        models: Union[StochasticModel, SME],
        target_critic: VValue,
    ):
        super().__init__()
        self.critics = critics

        if isinstance(actor, DeterministicPolicy):
            self.get_action = DeterministicPolicyAction(actor)
        elif isinstance(actor, StochasticPolicy):
            self.get_action = StochasticPolicyAction(actor)
        else:
            raise ValueError(f"Unsupported actor type '{type(actor)}'")

        if isinstance(models, StochasticModel):
            # Treat everything as if ensemble
            models = SME([models])
        self.models = models
        self.target_critic = target_critic

        self._loss_fn = nn.MSELoss()

    @property
    def model_samples(self) -> int:
        """Number of next states to sample from model."""
        return self._model_samples

    @model_samples.setter
    def model_samples(self, value: int):
        assert value > 0, "Number of model samples must be positive"
        self._model_samples = value

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self._env.initialized, (
            "Environment functions missing. "
            "Did you set reward and termination functions?"
        )
        obs = batch[SampleBatch.CUR_OBS]

        with torch.no_grad():
            action = self.get_action(obs)
            model, _ = self.sample_model()
            dist_params = model(obs, action)
            next_obs, _ = model.sample(dist_params, sample_shape=(self.model_samples,))

            reward = self._env.reward(obs, action, next_obs).mean(dim=0)
            next_values = self.target_critic(next_obs)
            dones = self._env.termination(obs, action, next_obs)
            next_values = torch.where(dones, torch.zeros_like(next_values), next_values)

            target = reward + self.gamma * next_values.mean(dim=0)

        values = self.critics(obs, action)
        loss = torch.stack([self._loss_fn(target, v) for v in values]).sum()

        stats = {"loss(critics)": loss.item()}
        stats.update(QLearningMixin.q_value_info(values))
        stats.update(dist_params_stats(dist_params, name="model"))
        return loss, stats



# pylint:disable=missing-module-docstring
from typing import Any, Optional, Tuple

import torch
from nnrl.nn.utils import perturb_params
from ray.rllib import SampleBatch
from ray.rllib.models.action_dist import ActionDistribution

from raylab.policy import TorchPolicy
from raylab.utils.param_noise import AdaptiveParamNoiseSpec, ddpg_distance_metric

from .base import Model
from .random_uniform import RandomUniform


class ParameterNoise(RandomUniform):
    """Adds adaptive parameter noise exploration schedule to a Policy.

    Expects `actor` attribute of `policy.module` to be an instance of
    `raylab.policy.modules.actor.policy.deterministic.DeterministicPolicy`.

    Args:
        param_noise_spec: Arguments for `AdaptiveParamNoiseSpec`.
    """

    def __init__(self, *args, param_noise_spec: dict = None, **kwargs):
        super().__init__(*args, **kwargs)
        param_noise_spec = param_noise_spec or {}
        self._param_noise_spec = AdaptiveParamNoiseSpec(**param_noise_spec)

    def get_exploration_action(
        self,
        *,
        action_distribution: ActionDistribution,
        timestep: int,
        explore: bool = True,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        if explore:
            if timestep < self._pure_exploration_steps:
                return super().get_exploration_action(
                    action_distribution=action_distribution,
                    timestep=timestep,
                    explore=explore,
                )
            return action_distribution.sample()
        return action_distribution.deterministic_sample()

    def on_episode_start(
        self,
        policy: TorchPolicy,
        *,
        environment: Any = None,
        episode: Any = None,
        tf_sess: Any = None,
    ):
        # pylint:disable=unused-argument
        perturb_params(
            policy.module.behavior,
            policy.module.actor,
            self._param_noise_spec.curr_stddev,
        )

    @torch.no_grad()
    def postprocess_trajectory(
        self, policy: TorchPolicy, sample_batch: SampleBatch, tf_sess: Any = None
    ):
        self.update_parameter_noise(policy, sample_batch)
        return sample_batch

    def get_info(self, sess=None) -> dict:
        return {"param_noise_stddev": self._param_noise_spec.curr_stddev}

    def update_parameter_noise(self, policy: TorchPolicy, sample_batch: SampleBatch):
        """Update parameter noise stddev given a batch from the perturbed policy."""
        module = policy.module
        cur_obs = policy.convert_to_tensor(sample_batch[SampleBatch.CUR_OBS])
        actions = policy.convert_to_tensor(sample_batch[SampleBatch.ACTIONS])

        noisy = module.actor.unsquash_action(actions)
        target = module.actor.unconstrained_action(cur_obs)
        noisy, target = map(lambda x: x.cpu().detach().numpy(), (noisy, target))

        distance = ddpg_distance_metric(noisy, target)
        self._param_noise_spec.adapt(distance)

    @classmethod
    def check_model_compat(cls, model: Model):
        assert (
            model is not None
        ), f"Need to pass the model to {cls} to check compatibility."
        actor, behavior = model.actor, model.behavior
        assert set(actor.parameters()).isdisjoint(set(behavior.parameters())), (
            "Target and behavior policy cannot share parameters in parameter "
            "noise exploration."
        )



"""SOP policy class using PyTorch."""
import torch
from nnrl.nn.critic import HardValue
from nnrl.nn.utils import update_polyak
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib.utils import override
from torch.nn.utils import clip_grad_norm_

from raylab.options import configure, option
from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import ActionDPG, DeterministicPolicyGradient, FittedQLearning
from raylab.policy.off_policy import OffPolicyMixin


@configure
@option("buffer_size", int(1e6))
@option("batch_size", 256)
@option("std_obs", False)
@option("improvement_steps", 1)
@option(
    "dpg_loss",
    "default",
    help="""
    Type of Deterministic Policy Gradient to use.

    'default' backpropagates Q-value gradients through the critic network.

    'acme' uses Acme's implementation which recovers DPG via a MSE loss between
    the actor's action and the action + Q-value gradient. Allows monitoring the
    magnitude of the action-value gradient.""",
)
@option(
    "dqda_clipping",
    None,
    help="""
    Optional value by which to clip the action gradients. Only used with
    dpg_loss='acme'.""",
)
@option(
    "clip_dqda_norm",
    False,
    help="""
    Whether to clip action grads by norm or value. Only used with
    dpg_loss='acme'.""",
)
@option("optimizer/actor", {"type": "Adam", "lr": 3e-4})
@option("optimizer/critics", {"type": "Adam", "lr": 3e-4})
@option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@option(
    "policy_delay",
    1,
    help="Update policy every this number of calls to `learn_on_batch`",
)
@option("module/type", "SOP")
@option("exploration_config/type", "raylab.utils.exploration.GaussianNoise")
@option("exploration_config/noise_stddev", 0.3)
@option("exploration_config/pure_exploration_steps", 10000)
class SOPTorchPolicy(OffPolicyMixin, TorchPolicy):
    """Streamlined Off-Policy policy in PyTorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapDeterministicPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._make_actor_loss()
        target_value = HardValue(self.module.target_actor, self.module.target_critics)
        self.loss_critic = FittedQLearning(self.module.critics, target_value)
        self.loss_critic.gamma = self.config["gamma"]
        self._grad_step = 0
        self._info = {}

        self.build_replay_buffer()

    def _make_actor_loss(self):
        if self.config["dpg_loss"] == "default":
            self.loss_actor = DeterministicPolicyGradient(
                self.module.actor,
                self.module.critics,
            )
        elif self.config["dpg_loss"] == "acme":
            self.loss_actor = ActionDPG(self.module.actor, self.module.critics)
            self.loss_actor.dqda_clipping = self.config["dqda_clipping"]
            self.loss_actor.clip_norm = self.config["clip_dqda_norm"]
        else:
            dpg_loss = self.config["dpg_loss"]
            raise ValueError(
                f"Invalid config for 'dpg_loss': {dpg_loss}."
                " Choose between 'default' and 'acme'"
            )

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        optimizers["actor"] = build_optimizer(self.module.actor, config["actor"])
        optimizers["critics"] = build_optimizer(self.module.critics, config["critics"])
        return optimizers

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict):
        self._grad_step += 1
        self._info["grad_steps"] = self._grad_step

        self._info.update(self._update_critic(batch))
        if self._grad_step % self.config["policy_delay"] == 0:
            self._info.update(self._update_policy(batch))

        critics, target_critics = self.module.critics, self.module.target_critics
        update_polyak(critics, target_critics, self.config["polyak"])
        return self._info.copy()

    def _update_critic(self, batch_tensors):
        with self.optimizers.optimize("critics"):
            loss, info = self.loss_critic(batch_tensors)
            loss.backward()

        info.update(self.extra_grad_info("critics"))
        return info

    def _update_policy(self, batch_tensors):
        with self.optimizers.optimize("actor"):
            loss, info = self.loss_actor(batch_tensors)
            loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component):
        """Return statistics right after components are updated."""
        params = getattr(self.module, component).parameters()
        return {f"grad_norm({component})": clip_grad_norm_(params, float("inf")).item()}

    @override(TorchPolicy)
    def get_weights(self):
        weights = super().get_weights()
        weights["grad_step"] = self._grad_step
        return weights

    @override(TorchPolicy)
    def set_weights(self, weights):
        self._grad_step = weights["grad_step"]
        super().set_weights(weights)



"""Utilities for dynamically importing modules."""
import os.path as osp
import pathlib
import sys
from importlib import import_module


def import_module_from_path(path):
    """Import a module given its path in the filesystem."""
    sys.path.append(osp.dirname(path))
    return import_module(pathlib.Path(path).stem)



"""Mixins to add environment mimicking functions."""
import inspect
import warnings
from typing import Optional, Union

from raylab import envs
from raylab.utils.types import DynamicsFn, RewardFn, TerminationFn


class EnvFnMixin:
    """Add methods to create and set environment functions.

    Allows user to choose methods for setting the reward, termination,
    and dynamics functions, which may vary by experiment setup.

    Attributes:
        reward_fn: callable mapping dynamics triplets to reward tensors
        termination_fn: callable mapping dynamics triplets to termination
            tensors
        dynamics_fn: callable mapping (state, action) pairs to (next-state,
            log-probability) pairs
    """

    reward_fn: Optional[RewardFn]
    termination_fn: Optional[TerminationFn]
    dynamics_fn: Optional[DynamicsFn]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reward_fn = None
        self.termination_fn = None
        self.dynamics_fn = None

    def set_reward_from_config(self):
        """Build and set the reward function from environment configurations."""
        env_id, env_config = self.config["env"], self.config["env_config"]
        self.reward_fn = envs.get_reward_fn(env_id, env_config)
        self._set_reward_hook()

    def set_reward_from_callable(self, function: RewardFn):
        """Set the reward function from an external callable.

        Args:
            function: A callable that reproduces the environment's reward
                function
        """
        self._check_not_instance_method(function)
        self.reward_fn = function
        self._set_reward_hook()

    def set_termination_from_config(self):
        """Build and set a termination function from environment configurations."""
        env_id, env_config = self.config["env"], self.config["env_config"]
        self.termination_fn = envs.get_termination_fn(env_id, env_config)
        self._set_termination_hook()

    def set_termination_from_callable(self, function: TerminationFn):
        """Set the termination function from an external callable.

        Args:
            function: A callable that reproduces the environment's termination
                function
        """
        self._check_not_instance_method(function)
        self.termination_fn = function
        self._set_termination_hook()

    def set_dynamics_from_callable(self, function: DynamicsFn):
        """Set the dynamics function from an external callable.

        Args:
            function: A callable that reproduces the environment's dynamics
                function
        """
        self._check_not_instance_method(function)
        self.dynamics_fn = function
        self._set_dynamics_hook()

    def _set_reward_hook(self):
        """Procedure to run upon setting the reward function.

        Subclasses should override this to do extra processing whenever
        `set_reward_from_callable` or `set_reward_from_config` is called.
        For example, the subclass may set up losses which assume access to
        environment functions.
        """

    def _set_termination_hook(self):
        """Procedure to run upon setting the termination function.

        Subclasses should override this to do extra processing whenever
        `set_termination_from_callable` or `set_termination_from_config` is
        called. For example, the subclass may set up losses which assume access
        to environment functions.
        """

    def _set_dynamics_hook(self):
        """Procedure to run upon setting the dynamics function.

        Subclasses should override this to do extra processing whenever
        `set_dynamics_from_callable` is called. For example, the subclass may
        set up losses which assume access to environment functions.
        """

    @staticmethod
    def _check_not_instance_method(env_fn: Union[RewardFn, TerminationFn]):
        if inspect.ismethod(env_fn):
            warnings.warn(
                f"{env_fn.__name__} function is a bound instance method of"
                f" {env_fn.__self__}. Prefer static environment functions"
                " with markovian states that are not bounded to a particular"
                " instance"
            )



"""SAC policy class using PyTorch."""
import torch
from nnrl.nn.critic import SoftValue
from nnrl.nn.utils import update_polyak
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib.utils import override
from torch import nn

from raylab.options import configure, option
from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.policy.losses import (
    FittedQLearning,
    MaximumEntropyDual,
    ReparameterizedSoftPG,
)
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options


@configure
@off_policy_options
@option(
    "target_entropy",
    None,
    help="""Target entropy for temperature parameter optimization.

    If 'auto', will use the heuristic provided in the SAC paper,
    H = -dim(A), where A is the action space

    If 'tf-agents', will use the TFAgents implementation,
    H = -dim(A) / 2, where A is the action space
    """,
)
@option("optimizer/actor", {"type": "Adam", "lr": 1e-3})
@option("optimizer/critics", {"type": "Adam", "lr": 1e-3})
@option("optimizer/alpha", {"type": "Adam", "lr": 1e-3})
@option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
@option("module", {"type": "SAC", "critic": {"double_q": True}}, override=True)
@option("exploration_config/pure_exploration_steps", 1000)
class SACTorchPolicy(OffPolicyMixin, TorchPolicy):
    """Soft Actor-Critic policy in PyTorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapStochasticPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)

        self._setup_actor_loss()
        self._setup_critic_loss()
        self._setup_alpha_loss()

        self.build_replay_buffer()

    def _setup_actor_loss(self):
        self.loss_actor = ReparameterizedSoftPG(
            actor=self.module.actor, critic=self.module.critics, alpha=self.module.alpha
        )

    def _setup_critic_loss(self):
        module = self.module
        soft_target = SoftValue(module.actor, module.target_critics, module.alpha)
        self.loss_critic = FittedQLearning(module.critics, soft_target)
        self.loss_critic.gamma = self.config["gamma"]

    def _setup_alpha_loss(self):
        action_size = self.action_space.shape[0]
        if self.config["target_entropy"] == "auto":
            target_entropy = -action_size
        elif self.config["target_entropy"] == "tf-agents":
            target_entropy = -action_size / 2
        else:
            target_entropy = self.config["target_entropy"]

        self.loss_alpha = MaximumEntropyDual(
            self.module.alpha, self.module.actor.sample, target_entropy
        )

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]

        components = "actor critics alpha".split()
        mapping = {
            name: build_optimizer(getattr(self.module, name), config[name])
            for name in components
        }

        optimizers.update(mapping)
        return optimizers

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict) -> dict:
        info = {}

        info.update(self._update_critic(batch))
        info.update(self._update_actor(batch))
        if self.config["target_entropy"] is not None:
            info.update(self._update_alpha(batch))

        update_polyak(
            self.module.critics, self.module.target_critics, self.config["polyak"]
        )
        return info

    def _update_critic(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("critics"):
            critic_loss, info = self.loss_critic(batch)
            critic_loss.backward()

        info.update(self.extra_grad_info("critics"))
        return info

    def _update_actor(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("actor"):
            actor_loss, info = self.loss_actor(batch)
            actor_loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    def _update_alpha(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("alpha"):
            alpha_loss, info = self.loss_alpha(batch)
            alpha_loss.backward()

        info.update(self.extra_grad_info("alpha"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component: str) -> dict:
        """Return statistics right after components are updated."""
        return {
            f"grad_norm({component})": nn.utils.clip_grad_norm_(
                getattr(self.module, component).parameters(), float("inf")
            ).item()
        }



# pylint:disable=missing-module-docstring
import click


@click.group("info")
def info_cli():
    """View information about an agent's config parameters."""


@info_cli.command("list")
@click.argument("agent")
@click.option("--key", type=str, default=None, help="Specific config to show info for.")
@click.option(
    "--separator",
    "-d",
    type=str,
    default="/",
    show_default=True,
    help="Separator for nested config keys.",
)
@click.option(
    "--rllib/--no-rllib",
    default=False,
    show_default=True,
    help="Whether to display RLlib's common config parameters and defaults. "
    "Warning: lots of parameters!",
)
@click.pass_context
def list_(ctx, agent, key, separator, rllib):
    """Retrieve and echo a help text for the given agent's config."""
    from raylab.agents.registry import AGENTS
    from raylab.options import UnknownOptionError

    cls = AGENTS[agent]()

    try:
        msg = cls.options.help(key, separator, with_rllib=rllib)
    except UnknownOptionError as err:
        click.echo(err)
        click.echo(f"Exception raised when querying agent {agent} options for help.")
        ctx.exit()

    click.echo(msg)



"""Standard PyTorch optimizer interface for TorchPolicies."""
import contextlib
from collections import OrderedDict
from collections.abc import MutableMapping
from typing import Iterator

from torch.optim import Optimizer


class OptimizerCollection(MutableMapping):
    """A collection of PyTorch `Optimizer`s with names."""

    def __init__(self):
        self._optimizers = OrderedDict()

    def __setitem__(self, key: str, value: Optimizer):
        """Adds an optimizer to the collection.

        Args:
            key: the name of the optimizer
            value: the optimizer instance

        Raises:
            ValueError: if the value to insert is not an optimizer
        """
        assert key not in self._optimizers, f"'{key}' optimizer already in collection"
        if not isinstance(value, Optimizer):
            raise ValueError("'{type(value).__name__}' is not an Optimizer instance")

        self._optimizers[key] = value

    def __getitem__(self, key: str) -> Optimizer:
        """Gets an optimizer from the collection.

        Args:
            key: the name of the optimizer

        Returns:
            The optimizer instance
        """
        if key not in self._optimizers:
            raise KeyError(f"'{key}' optimizer not in collection")
        return self._optimizers[key]

    def __delitem__(self, key: str):
        """Removes an optimizer from the collection.

        Args:
            key: the name of the optimizer
        """
        if key not in self._optimizers:
            raise KeyError(f"'{key}' optimizer not in collection")
        del self._optimizers[key]

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator over the optimizer names in the collection."""
        return iter(self._optimizers.keys())

    def __len__(self) -> int:
        """Returns the number of optimizers in the collection."""
        return len(self._optimizers)

    @contextlib.contextmanager
    def optimize(self, name: str):
        """Nullify grads before context and step optimizer afterwards."""
        optimizer = self[name]
        optimizer.zero_grad()
        yield
        optimizer.step()

    def state_dict(self) -> dict:
        """Returns the state of each optimizer in the collection."""
        return {k: v.state_dict() for k, v in self.items()}

    def load_state_dict(self, state_dict: dict):
        """Loads the state of each optimizer.

        Args:
            state_dict: optimizer collection state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        for name, optim in self._optimizers.items():
            optim.load_state_dict(state_dict[name])



"""Global registry of rewards, terminations, and raylab modules.

Mimics Tune's registry in `ray.tune.registry`. Such as registry is needed so
that users can create custom reward and termination functions and use them in
a distributed setting.
"""
import ray
import ray.cloudpickle as pickle
from ray.experimental.internal_kv import _internal_kv_initialized
from ray.tune.registry import _Registry as _TuneRegistry

RAYLAB_REWARD = "raylab_reward"
RAYLAB_TERMINATION = "raylab_termination"
RAYLAB_MODULE = "raylab_module"
KNOWN_CATEGORIES = {
    RAYLAB_REWARD,
    RAYLAB_TERMINATION,
    RAYLAB_MODULE,
}


class _RaylabRegistry(_TuneRegistry):
    # pylint:disable=missing-docstring
    def register(self, category, key, value):
        if category not in KNOWN_CATEGORIES:
            from ray.tune import TuneError

            raise TuneError(
                "Unknown category {} not among {}".format(category, KNOWN_CATEGORIES)
            )
        self._to_flush[(category, key)] = pickle.dumps(value)
        if _internal_kv_initialized():
            self.flush_values()


_raylab_registry = _RaylabRegistry()
# pylint:disable=protected-access
ray.worker._post_init_hooks.append(_raylab_registry.flush_values)



"""Continuous Q-Learning with Normalized Advantage Functions."""
from raylab.agents import Trainer
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import NAFTorchPolicy


@configure
@OffPolicyMixin.add_options
class NAFTrainer(OffPolicyMixin, Trainer):
    """Single agent trainer for NAF."""

    _name = "NAF"
    _policy_class = NAFTorchPolicy
